"""Test Jac language generally."""

from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCase


class JacBugTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_impl_match_confusion_issue(self) -> None:
        """Basic test for symtable support for inheritance."""
        (mypass := JacProgram()).compile(
            self.fixture_abs_path("impl_match_confused.jac"),
        )
        self.assertEqual(len(mypass.errors_had), 1)



"""Test Jac language generally."""

import io
import os
import sys
import unittest
import sysconfig
import tempfile
from unittest.mock import patch

from jaclang import JacMachineInterface as Jac, JacMachine
from jaclang.cli import cli
from jaclang.compiler.program import JacProgram
from jaclang.utils.lang_tools import AstTool
from jaclang.utils.test import TestCase


@unittest.skip("Skipping typecheck tests")
class JacTypeCheckTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        self.mach = JacMachine(self.fixture_abs_path("./"))
        Jac.attach_program(
            self.mach,
            JacProgram(),
        )
        return super().setUp()

    def tearDown(self) -> None:
        """Tear down test."""
        return super().tearDown()

    def test_multi_dim_arr_slice(self) -> None:
        """Parse micro jac file."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.tool(
            "ir",
            [
                "ast",
                self.fixture_abs_path("multi_dim_array_split.jac"),
            ],
        )
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()

        expected_outputs = [
            "+-- AtomTrailer - Type: builtins.list[builtins.int]",
            "    +-- Name - arr - Type: builtins.list[builtins.list[builtins.int]],  SymbolTable: list",
            "+-- IndexSlice - [IndexSlice] - Type: builtins.list[builtins.list[builtins.int]],  SymbolTable: None",
            "        +-- Token - [, ",
            "        +-- Int - 1 - Type: Literal[1]?,  SymbolTable: None",
            "        +-- Token - :, ",
            "        +-- Int - 3 - Type: Literal[3]?,  SymbolTable: None",
            "        +-- Token - ,, ",
            "        +-- Int - 1 - Type: Literal[1]?,  SymbolTable: None",
            "        +-- Token - :, ",
            "        +-- Token - :, ",
            "        +-- Int - 2 - Type: Literal[2]?,  SymbolTable: None",
            "        +-- Token - ], ",
        ]

        for expected in expected_outputs:
            self.assertIn(expected, stdout_value)

    def test_needs_import_1(self) -> None:
        """Test py ast to Jac ast conversion output."""
        file_name = self.fixture_abs_path("pyfunc_1.py")

        from jaclang.compiler.passes.main.pyast_load_pass import PyastBuildPass
        import ast as py_ast
        import jaclang.compiler.unitree as uni

        with open(file_name, "r") as f:
            file_source = f.read()
            parsed_ast = py_ast.parse(file_source)
            try:
                py_ast_build_pass = PyastBuildPass(
                    ir_in=uni.PythonModuleAst(
                        parsed_ast, orig_src=uni.Source(file_source, file_name)
                    ),
                    prog=JacProgram(),
                ).ir_out
            except Exception as e:
                return f"Error While Jac to Py AST conversion: {e}"

        (prog := JacProgram()).build(
            use_str=py_ast_build_pass.unparse(),
            file_path=file_name[:-3] + ".jac",
        )

        archetype_count = 0
        for mod in prog.mod.hub.values():
            if mod.name == "builtins":
                continue
            archetype_count += len(mod.get_all_sub_nodes(uni.Archetype))

        self.assertEqual(archetype_count, 24)
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import(
            self.mach, "needs_import_1", base_path=self.fixture_abs_path("./")
        )
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("pyfunc_1 imported", stdout_value)

    def test_needs_import_2(self) -> None:
        """Test py ast to Jac ast conversion output."""
        file_name = self.fixture_abs_path("pyfunc_2.py")

        from jaclang.compiler.passes.main.pyast_load_pass import PyastBuildPass
        import ast as py_ast
        import jaclang.compiler.unitree as uni

        with open(file_name, "r") as f:
            file_source = f.read()
            parsed_ast = py_ast.parse(file_source)
            try:
                py_ast_build_pass = PyastBuildPass(
                    ir_in=uni.PythonModuleAst(
                        parsed_ast,
                        orig_src=uni.Source(file_source, file_name),
                    ),
                    prog=JacProgram(),
                ).ir_out
            except Exception as e:
                return f"Error While Jac to Py AST conversion: {e}"

            (prog := JacProgram()).build(
                use_str=py_ast_build_pass.unparse(),
                file_path=file_name[:-3] + ".jac",
            )

        archetype_count = 0
        for mod in prog.mod.hub.values():
            if mod.name == "builtins":
                continue
            archetype_count += len(mod.get_all_sub_nodes(uni.Archetype))

        self.assertEqual(archetype_count, 30)  # Because of the Archetype from math
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import(
            self.mach, "needs_import_2", base_path=self.fixture_abs_path("./")
        )
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("pyfunc_2 imported", stdout_value)
        self.assertEqual(stdout_value.count("<class 'bytes'>"), 3)

    def test_needs_import_3(
        self,
    ) -> None:  # TODO : Pyfunc_3 has a bug in conversion in matchmapping node
        """Test py ast to Jac ast conversion output."""
        file_name = self.fixture_abs_path("pyfunc_3.py")
        import jaclang.compiler.unitree as uni

        with open(file_name, "r") as f:
            file_source = f.read()
        (prog := JacProgram()).build(use_str=file_source, file_path=file_name)

        archetype_count = sum(
            len(mod.get_all_sub_nodes(uni.Archetype))
            for mod in prog.mod.hub.values()
            if mod.name != "builtins"
        )
        self.assertEqual(
            archetype_count, 58
        )  # Fixed duplication of 'case' module (previously included 3 times, added 20 extra Archetypes; 75 → 55)
        builtin_mod = next(
            (mod for name, mod in prog.mod.hub.items() if "builtins" in name),
            None,
        )
        self.assertEqual(len(builtin_mod.get_all_sub_nodes(uni.Archetype)), 109)
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import(
            self.mach, "needs_import_3", base_path=self.fixture_abs_path("./")
        )
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("pyfunc_3 imported", stdout_value)

    def test_type_fuse_expr(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.tool(
            "ir",
            [
                "ast",
                self.examples_abs_path("reference/collection_values.jac"),
            ],
        )

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn(
            "builtins.dict[builtins.int, builtins.int]",
            stdout_value,
        )
        self.assertIn(
            "typing.Generator[builtins.int, None, None]",
            stdout_value,
        )

    def test_random_check(self) -> None:
        """Test py ast to Jac ast conversion output."""
        from jaclang.settings import settings

        module_paths = ["random", "ast"]
        for module_path in module_paths:
            stdlib_dir = sysconfig.get_paths()["stdlib"]
            file_path = os.path.join(
                stdlib_dir,
                module_path + ".py",
            )
            settings.print_py_raised_ast = True
            with open(file_path) as f:
                file_source = f.read()
            ir = JacProgram().build(
                use_str=file_source,
                file_path=file_path,
            )
            gen_ast = ir.pp()
            if module_path == "random":
                self.assertIn("ModulePath - statistics -", gen_ast)
            else:
                self.assertIn("+-- Name - NodeTransformer - Type: No", gen_ast)

    def test_deep_convert(self) -> None:
        """Test py ast to Jac ast conversion output."""
        file_name = self.fixture_abs_path("pyfunc_1.py")

        import jaclang.compiler.unitree as uni
        from jaclang.settings import settings

        settings.print_py_raised_ast = True
        with open(file_name, "r") as f:
            file_source = f.read()
        ir = (prog := JacProgram()).build(use_str=file_source, file_path=file_name)
        jac_ast = ir.pp()
        self.assertIn(" |   +-- String - 'Loop completed normally{}'", jac_ast)
        sub_node_list_count = 0
        for i in prog.mod.hub.values():
            if i.name == "builtins":
                continue
            sub_node_list_count += len(i.get_all_sub_nodes(uni.SubNodeList))
        self.assertEqual(sub_node_list_count, 623)
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import(self.mach, "deep_convert", base_path=self.fixture_abs_path("./"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Deep convo is imported", stdout_value)

    def test_ds_type_check_pass(self) -> None:
        """Test conn assign on edges."""
        (mypass := JacProgram()).build(
            self.examples_abs_path("micro/simple_walk.jac"),
        )
        self.assertEqual(len(mypass.errors_had), 0)
        self.assertEqual(len(mypass.warnings_had), 0)

    def test_ds_type_check_pass2(self) -> None:
        """Test conn assign on edges."""
        (mypass := JacProgram()).build(
            self.examples_abs_path("guess_game/guess_game5.jac"),
        )
        self.assertEqual(len(mypass.errors_had), 0)
        self.assertEqual(len(mypass.warnings_had), 0)

    def test_circle_override1_type_check_pass(self) -> None:
        """Test conn assign on edges."""
        (mypass := JacProgram()).build(
            self.examples_abs_path("manual_code/circle.jac"),
        )
        self.assertEqual(len(mypass.errors_had), 0)
        # FIXME: Figure out what to do with warning.
        # self.assertEqual(len(mypass.warnings_had), 0)

    def test_expr_types(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('expr_type.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()

        self.assertRegex(
            stdout_value, r"4\:9 \- 4\:14.*BinaryExpr \- Type\: builtins.int"
        )
        self.assertRegex(
            stdout_value, r"7\:9 \- 7\:17.*FuncCall \- Type\: builtins.float"
        )
        self.assertRegex(
            stdout_value, r"9\:6 \- 9\:11.*CompareExpr \- Type\: builtins.bool"
        )
        self.assertRegex(
            stdout_value, r"10\:6 - 10\:15.*BinaryExpr \- Type\: builtins.str"
        )
        self.assertRegex(
            stdout_value, r"11\:5 \- 11\:13.*AtomTrailer \- Type\: builtins.int"
        )
        self.assertRegex(
            stdout_value, r"12\:5 \- 12\:14.*UnaryExpr \- Type\: builtins.bool"
        )
        self.assertRegex(
            stdout_value, r"13\:5 \- 13\:25.*IfElseExpr \- Type\: Literal\['a']\?"
        )
        self.assertRegex(
            stdout_value,
            r"14\:5 \- 14\:27.*ListCompr - \[ListCompr] \- Type\: builtins.list\[builtins.int]",
        )

    def test_type_check(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.check(f"{self.fixture_abs_path('game1.jac')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Errors: 0, Warnings: 1", stdout_value)

    def test_type_errors(self) -> None:
        """Basic test for pass."""
        (type_checked := JacProgram()).build(
            file_path=self.fixture_abs_path("func.jac"),
        )

        errs = "\n".join([i.msg for i in type_checked.warnings_had])
        files = "\n".join([i.loc.mod_path for i in type_checked.warnings_had])

        for i in [
            "func2.jac",
            "func.jac",
            '(got "int", expected "str")',
            '(got "str", expected "int")',
        ]:
            self.assertIn(i, errs + files)

    def test_imported_module_typecheck(self) -> None:
        """Basic test for pass."""
        (type_checked := JacProgram()).build(
            file_path=self.fixture_abs_path("game1.jac"),
        )

        errs = "\n".join([i.msg for i in type_checked.warnings_had])
        files = "\n".join([i.loc.mod_path for i in type_checked.warnings_had])

        for i in [
            'Argument 2 to "is_pressed" of "Button" has incompatible type "int"; expected "str"',
        ]:
            self.assertIn(i, errs + files)

    def test_type_coverage(self) -> None:
        """Testing for type info coverage in sym_tab via ast."""
        out = AstTool().ir(["ast", f"{self.fixture_abs_path('type_info.jac')}"])
        lis = [
            "introduce - Type: None",
            "25:30 - 25:37\t        |   |   |           +-- Name - species - Type: None",  # AtomTrailer
        ]
        self.assertIn("HasVar - species - Type: builtins.str", out)
        self.assertIn("myDog - Type: type_info.Dog", out)
        self.assertIn("Body - Type: type_info.Dog.Body", out)
        self.assertEqual(out.count("Type: builtins.str"), 35)
        for i in lis:
            self.assertNotIn(i, out)

    def test_data_spatial_type_info(self) -> None:
        """Testing for type info for dataspatial constructs."""
        out = AstTool().ir(
            ["ast", f"{self.fixture_abs_path('data_spatial_types.jac')}"]
        )
        self.assertRegex(
            out,
            r"129:22 - 129:26.*SpecialVarRef - root \- Type\: jaclang.runtimelib.archetype.Root",
        )
        self.assertRegex(out, r"129:11 - 129:27.*FuncCall \- Type\: builtins\.str")
        self.assertRegex(
            out,
            r"129:13 - 129:21.*Name \- node_dot \- Type\: builtins.str",
        )

        self.assertRegex(
            out,
            r"128:5 - 128:25.*BinaryExpr \- Type\: "
            + r"Union\[jaclang.runtimelib.archetype.WalkerArchetype, concurrent.futures._base.Future\[Any\]\]",
        )

        self.assertRegex(
            out,
            r"48:11 - 48:28.*EdgeRefTrailer \- Type\: builtins.list\[jaclang.runtimelib.archetype.Archetype\]",
        )

        self.assertRegex(out, r"24:5 - 24:25.*BinaryExpr \- Type\: builtins.bool", out)

    def test_type_info(self) -> None:
        """Testing for type info inside the ast tool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.tool("ir", ["ast", f"{self.fixture_abs_path('type_info.jac')}"])
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertEqual(stdout_value.count("type_info.ServerWrapper"), 5)
        self.assertEqual(stdout_value.count("builtins.int"), 3)
        self.assertEqual(stdout_value.count("builtins.str"), 10)
        self.assertIn("Literal['test_server']", stdout_value)
        self.assertIn("Literal['1']", stdout_value)

    def test_base_class_complex_expr(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        # settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool(
            "ir", ["ast", f"{self.fixture_abs_path('base_class_complex_expr.jac')}"]
        )

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"36\:9 \- 36\:13.*Name \- Kiwi \- Type\: base_class_complex_expr.Kiwi,  SymbolTable\: Kiwi",
        )

    def test_import_all(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('import_all.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"6\:25 - 6\:30.*Name - floor -.*SymbolPath: math.floor",
        )
        self.assertRegex(
            stdout_value,
            r"5\:25 - 5\:27.*Name - pi -.*SymbolPath: math.pi",
        )

    def test_import_mod_abs_path(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('import.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue().replace("\\", "/")
        self.assertRegex(
            stdout_value,
            r"1.*ModulePath - os - abs_path\:.*typeshed/stdlib/os/__init__.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"2.*ModulePath - sys - abs_path\:.*typeshed/stdlib/sys/__init__.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"3.*ModulePath - pyfunc - abs_path\:.*fixtures/pyfunc.py",
        )
        self.assertRegex(
            stdout_value,
            r"4.*ModulePath - pygame_mock - abs_path\:.*fixtures/pygame_mock/inner/__init__.py",
        )
        self.assertRegex(
            stdout_value,
            r"6.*ModulePath - math - abs_path\:.*typeshed/stdlib/math.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"7.*ModulePath - argparse - abs_path\:.*typeshed/stdlib/argparse.pyi",
        )
        self.assertRegex(
            stdout_value,
            r"8.*ModulePath - pygame_mock - abs_path\:.*fixtures/pygame_mock/__init__.py",
        )
        self.assertRegex(
            stdout_value,
            r"8.*ModuleItem - color - abs_path\:.*fixtures/pygame_mock/color.py",
        )
        self.assertRegex(
            stdout_value,
            r"8.*ModuleItem - display - abs_path\:.*fixtures/pygame_mock/display.py",
        )

    def test_sub_class_symbol_table_fix_1(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('base_class1.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"10:7 - 10:12.*Name - start - Type.*SymbolPath: base_class1.B.start",
        )

    def test_sub_class_symbol_table_fix_2(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('base_class2.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"10:7 - 10:12.*Name - start - Type.*SymbolPath: base_class2.B.start",
        )



"""Module Import Resolution Pass for the Jac compiler.

This pass handles the static resolution and loading of imported modules by:

1. Identifying import statements in the source code
2. Resolving module paths (both relative and absolute)
3. Loading and parsing the imported modules
4. Handling both Jac and Python imports with appropriate strategies
5. Managing import dependencies and preventing circular imports
6. Supporting various import styles:
   - Direct imports (import x)
   - From imports (from x import y)
   - Star imports (from x import *)
   - Aliased imports (import x as y)

The pass runs early in the compilation pipeline to ensure all symbols from imported
modules are available for subsequent passes like symbol table building and type checking.
"""

import ast as py_ast
import os
from typing import Optional


import jaclang.compiler.unitree as uni
from jaclang.compiler.passes import Transform, UniPass
from jaclang.compiler.passes.main import SymTabBuildPass
from jaclang.utils.log import logging


logger = logging.getLogger(__name__)


# TODO: This pass finds imports dependencies, parses them, and adds them to
# JacProgram's table, then table calls again if needed, should rename
class JacImportDepsPass(Transform[uni.Module, uni.Module]):
    """Jac statically imports Jac modules."""

    def pre_transform(self) -> None:
        """Initialize the JacImportPass."""
        super().pre_transform()
        self.last_imported: list[uni.Module] = []

    def transform(self, ir_in: uni.Module) -> uni.Module:
        """Run Importer."""
        # Add the current module to last_imported to start the import process
        self.last_imported.append(ir_in)

        # Process imports until no more imported modules to process
        while self.last_imported:
            current_module = self.last_imported.pop(0)
            all_imports = UniPass.get_all_sub_nodes(current_module, uni.ModulePath)
            for i in all_imports:
                self.process_import(i)

        return ir_in

    def process_import(self, i: uni.ModulePath) -> None:
        """Process an import."""
        imp_node = i.parent_of_type(uni.Import)
        if imp_node.is_jac:
            self.import_jac_module(node=i)

    def import_jac_module(self, node: uni.ModulePath) -> None:
        """Import a module."""
        target = node.resolve_relative_path()
        # If the module is a package (dir)
        if os.path.isdir(target):
            self.load_mod(self.import_jac_mod_from_dir(target))
            import_node = node.parent_of_type(uni.Import)
            # And the import is a from import and I am the from module
            if node == import_node.from_loc:
                # Import all from items as modules or packages
                for i in import_node.items:
                    if isinstance(i, uni.ModuleItem):
                        from_mod_target = node.resolve_relative_path(i.name.value)
                        # If package
                        if os.path.isdir(from_mod_target):
                            self.load_mod(self.import_jac_mod_from_dir(from_mod_target))
                        # Else module
                        else:
                            if from_mod_target in self.prog.mod.hub:
                                return
                            self.load_mod(self.prog.compile(file_path=from_mod_target))
        else:
            if target in self.prog.mod.hub:
                return
            self.load_mod(self.prog.compile(file_path=target))

    def load_mod(self, mod: uni.Module) -> None:
        """Attach a module to a node."""
        self.prog.mod.hub[mod.loc.mod_path] = mod
        self.last_imported.append(mod)

    # TODO: Refactor this to a function for impl and function for test

    def import_jac_mod_from_dir(self, target: str) -> uni.Module:
        """Import a module from a directory."""
        jac_init_path = os.path.join(target, "__init__.jac")
        if os.path.exists(jac_init_path):
            if jac_init_path in self.prog.mod.hub:
                return self.prog.mod.hub[jac_init_path]
            return self.prog.compile(file_path=jac_init_path)
        elif os.path.exists(py_init_path := os.path.join(target, "__init__.py")):
            with open(py_init_path, "r") as f:
                file_source = f.read()
                mod = uni.Module.make_stub(
                    inject_name=target.split(os.path.sep)[-1],
                    inject_src=uni.Source(file_source, py_init_path),
                )
                self.prog.mod.hub[py_init_path] = mod
                return mod
        else:
            return uni.Module.make_stub(
                inject_name=target.split(os.path.sep)[-1],
                inject_src=uni.Source("", target),
            )


class PyImportDepsPass(JacImportDepsPass):
    """Jac statically imports Python modules."""

    def pre_transform(self) -> None:
        """Initialize the PyImportPass."""
        self.import_from_build_list: list[tuple[uni.Import, uni.Module]] = []

    def transform(self, ir_in: uni.Module) -> uni.Module:
        """Run Importer."""
        self.__load_builtins()
        self.import_from_build_list = []

        # Add all modules from the program hub to last_imported to process their imports
        self.last_imported = list(self.prog.mod.hub.values())

        # Process imports until no more imported modules to process
        while self.last_imported:
            current_module = self.last_imported.pop(0)
            all_imports = UniPass.get_all_sub_nodes(current_module, uni.ModulePath)
            for i in all_imports:
                self.process_import(i)

        return ir_in

    def process_import(self, i: uni.ModulePath) -> None:
        """Process an import."""
        # Process import is orginally implemented to handle ModulePath in Jaclang
        # This won't work with py imports as this will fail to import stuff in form of
        #      from a import b
        #      from a import (c, d, e)
        # Solution to that is to get the import node and check the from loc `then`
        # handle it based on if there a from loc or not
        imp_node = i.parent_of_type(uni.Import)

        if imp_node.is_py:
            if imp_node.from_loc:
                msg = "Processing import from node at href="
                msg += uni.Module.get_href_path(imp_node)
                msg += f' path="{imp_node.loc.mod_path}, {imp_node.loc}"'
                self.__process_import_from(imp_node)
            else:
                msg = "Processing import node at href="
                msg += uni.Module.get_href_path(imp_node)
                msg += f' path="{imp_node.loc.mod_path}, {imp_node.loc}"'
                self.__process_import(imp_node)

    def __process_import_from(self, imp_node: uni.Import) -> None:
        """Process imports in the form of `from X import I`."""
        assert isinstance(imp_node.from_loc, uni.ModulePath)

        # Attempt to import the Python module X and process it
        imported_mod = self.__import_py_module(
            parent_node_path=uni.Module.get_href_path(imp_node),
            mod_path=imp_node.from_loc.dot_path_str,
        )

        if imported_mod:
            # Cyclic imports will happen in case of import sys
            # sys stub file imports sys module which means that we need
            # to import sys stub file again in the sys stub file and so on
            # This can be detected by iterating over all the parents and make sure
            # that the parent is in another file than the imported module
            if self.__check_cyclic_imports(imp_node, imported_mod):
                return

            if imported_mod.name == "builtins":
                return

            msg = f"\tRegistering module:{imported_mod.name} to "
            msg += f"import_from handling with {imp_node.loc.mod_path}:{imp_node.loc}"

            self.load_mod(imported_mod)
            self.import_from_build_list.append((imp_node, imported_mod))
            SymTabBuildPass(ir_in=imported_mod, prog=self.prog)

    def __process_import(self, imp_node: uni.Import) -> None:
        """Process the imports in form of `import X`."""
        # Expected that each ImportStatement will import one item
        # In case of this assertion fired then we need to revisit this item
        assert len(imp_node.items) == 1
        imported_item = imp_node.items[0]
        assert isinstance(imported_item, uni.ModulePath)

        imported_mod = self.__import_py_module(
            parent_node_path=uni.Module.get_href_path(imported_item),
            mod_path=imported_item.dot_path_str,
            imported_mod_name=(
                # TODO: Check this replace
                imported_item.dot_path_str.replace(".", "")
                if not imported_item.alias
                else imported_item.alias.sym_name
            ),
        )
        if imported_mod:
            if (
                self.__check_cyclic_imports(imp_node, imported_mod)
                or imported_mod.name == "builtins"
            ):
                return
            self.load_mod(imported_mod)

            if imp_node.is_absorb:
                msg = f"\tRegistering module:{imported_mod.name} to "
                msg += f"import_from (import all) handling with {imp_node.loc.mod_path}:{imp_node.loc}"

                self.import_from_build_list.append((imp_node, imported_mod))
            SymTabBuildPass(ir_in=imported_mod, prog=self.prog)

    def __import_py_module(
        self,
        parent_node_path: str,
        mod_path: str,
        imported_mod_name: Optional[str] = None,
    ) -> Optional[uni.Module]:
        """Import a python module."""
        from jaclang.compiler.passes.main import PyastBuildPass

        python_raise_map = self.prog.py_raise_map
        file_to_raise: Optional[str] = None

        if mod_path in python_raise_map:
            file_to_raise = python_raise_map[mod_path]
        else:
            # TODO: Is it fine to use imported_mod_name or get it from mod_path?
            resolved_mod_path = f"{parent_node_path}.{mod_path}"
            resolved_mod_path = resolved_mod_path.replace("..", ".")
            resolved_mod_path = resolved_mod_path.replace(
                f"{list(self.prog.mod.hub.values())[0]}.", ""
            )
            file_to_raise = python_raise_map.get(resolved_mod_path)

        if file_to_raise is None:
            return None

        try:
            if file_to_raise in {None, "built-in", "frozen"}:
                return None

            with open(file_to_raise, "r", encoding="utf-8") as f:
                file_source = f.read()
                mod = PyastBuildPass(
                    ir_in=uni.PythonModuleAst(
                        py_ast.parse(file_source),
                        orig_src=uni.Source(file_source, file_to_raise),
                    ),
                    prog=self.prog,
                ).ir_out

            if mod:
                mod.name = imported_mod_name if imported_mod_name else mod.name
                if mod.name == "__init__":
                    # (thakee): This needs to be re-done after implementing path handling properly.
                    mod_name = mod.loc.mod_path.split(os.path.sep)[-2]
                    mod.name = mod_name
                    mod.scope_name = mod_name
                mod.is_raised_from_py = True
                return mod
            else:
                raise self.ice(f"\tFailed to import python module {mod_path}")

        except Exception as e:
            self.log_error(f"\tFailed to import python module {mod_path}")
            raise e

    def __load_builtins(self) -> None:
        """Load Python builtins using introspection."""
        import builtins
        import inspect
        from jaclang.compiler.passes.main import PyastBuildPass

        # Python constants that cannot be assigned to
        constants = {"True", "False", "None", "NotImplemented", "Ellipsis", "..."}

        # Create a synthetic source with all builtins
        builtin_items = []
        for name in dir(builtins):
            if name in constants:
                # Skip constants as they're built into Python
                continue

            if not name.startswith("_") or name in ("__import__", "__build_class__"):
                obj = getattr(builtins, name)
                # Generate appropriate stub definitions based on obj type
                if inspect.isclass(obj):
                    builtin_items.append(f"class {name}: ...")
                elif inspect.isfunction(obj) or inspect.isbuiltin(obj):
                    # Try to get signature safely, use generic signature if it fails
                    try:
                        sig = inspect.signature(obj) if callable(obj) else "()"
                        builtin_items.append(f"def {name}{sig}: ...")
                    except (ValueError, TypeError):
                        builtin_items.append(f"def {name}(*args, **kwargs): ...")
                else:
                    # For variables that are not constants
                    builtin_items.append(f"{name} = None")

        file_source = "\n".join(builtin_items)
        mod = PyastBuildPass(
            ir_in=uni.PythonModuleAst(
                py_ast.parse(file_source),
                orig_src=uni.Source(file_source, "builtins"),
            ),
            prog=self.prog,
        ).ir_out
        SymTabBuildPass(ir_in=mod, prog=self.prog)
        self.prog.mod.hub["builtins"] = mod
        mod.is_raised_from_py = True

    def annex_impl(self, node: uni.Module) -> None:
        """Annex impl and test modules."""
        return None

    def __handle_different_site_packages(self, mod_path: str) -> str:
        if "site-packages" in mod_path:
            mod_path = mod_path[mod_path.index("site-packages") :]
        return mod_path

    def __check_cyclic_imports(
        self, imp_node: uni.UniNode, imported_module: uni.Module
    ) -> bool:
        """Check cyclic imports that might happen."""
        # Example of cyclic imports is import os
        # In the os stub file it imports the real os module which will cause os
        # stub to be raised again and so on
        # Another example is numpy. numpy init file imports multidim array file
        # which imports again more items from numpy and so on.
        imp_node_file = self.__handle_different_site_packages(imp_node.loc.mod_path)
        imported_module_file = self.__handle_different_site_packages(
            imported_module.loc.mod_path
        )
        if imp_node_file == imported_module_file:
            return True

        parent: Optional[uni.UniNode] = imp_node.parent
        while parent is not None:
            parent_file = self.__handle_different_site_packages(parent.loc.mod_path)
            if parent_file == imported_module_file:
                return True
            else:
                parent = parent.find_parent_of_type(uni.Module)

        return False



"""Utility functions and classes for Jac compilation toolchain."""

import dis
import marshal
import os
import pdb
import re
from traceback import TracebackException


def pascal_to_snake(pascal_string: str) -> str:
    """Convert pascal case to snake case."""
    snake_string = re.sub(r"(?<!^)(?=[A-Z])", "_", pascal_string).lower()
    return snake_string


def heading_to_snake(heading: str) -> str:
    """Convert string to snakecase including replacing(/ ,- )."""
    return heading.strip().replace("-", "_").replace("/", "_").replace(" ", "_").lower()


def add_line_numbers(s: str) -> str:
    """Add line numbers to a string."""
    lines = s.split("\n")
    return "\n".join(f"{i + 1}: \t{line}" for i, line in enumerate(lines))


def clip_code_section(s: str, target_line: int, line_range: int) -> str:
    """Clip a section of code and highlight target line."""
    lines = s.split("\n")
    start = max(0, target_line - line_range - 1)
    end = min(target_line + line_range, len(lines))

    result = []
    for i in range(start, end):
        line = lines[i]
        if i == target_line - 1:
            line = "*" + line
        result.append(line)
    return "\n".join(result)


def get_uni_nodes_as_snake_case() -> list[str]:
    """Get all AST nodes as snake case."""
    import inspect
    import sys
    import jaclang.compiler.unitree as uni

    module_name = uni.__name__
    module = sys.modules[module_name]

    # Retrieve the source code of the module
    source_code = inspect.getsource(module)

    classes = inspect.getmembers(module, inspect.isclass)
    uni_node_classes = [cls for _, cls in classes if issubclass(cls, uni.UniNode)]

    ordered_classes = sorted(
        uni_node_classes, key=lambda cls: source_code.find(f"class {cls.__name__}")
    )
    snake_names = []
    for cls in ordered_classes:
        class_name = cls.__name__
        snake_names.append(pascal_to_snake(class_name))
    return snake_names


def extract_headings(file_path: str) -> dict[str, tuple[int, int]]:
    """Extract headings of contetnts in Jac grammer."""
    with open(file_path, "r") as file:
        lines = file.readlines()
    headings = {}
    current_heading = None
    start_line = 0
    for idx, line in enumerate(lines, start=1):
        line = line.strip().removesuffix(".")
        if line.startswith("// [Heading]:"):
            if current_heading is not None:
                headings[current_heading] = (
                    start_line,
                    idx - 2,
                )  # Subtract 1 to get the correct end line
            current_heading = line.removeprefix("// [Heading]:")
            start_line = idx + 1
    # Add the last heading
    if current_heading is not None:
        headings[current_heading] = (start_line, len(lines))
    return headings


def auto_generate_refs() -> str:
    """Auto generate lang reference for docs."""
    file_path = os.path.join(
        os.path.split(os.path.dirname(__file__))[0], "../jaclang/compiler/jac.lark"
    )
    result = extract_headings(file_path)
    md_str = '# Jac Language Reference\n\n--8<-- "jac/examples/reference/introduction.md"\n\n'
    for heading, lines in result.items():
        heading = heading.strip()
        heading_snakecase = heading_to_snake(heading)
        content = (
            f'## {heading}\n**Code Example**\n!!! example "Runnable Example in Jac and JacLib"\n'
            '    === "Try it!"\n        <div class="code-block">\n'
            "        ```jac\n"
            f'        --8<-- "jac/examples/reference/{heading_snakecase}.jac"\n'
            "        ```\n"
            "        </div>\n"
            '    === "Jac"\n        ```jac linenums="1"\n'
            f'        --8<-- "jac/examples/reference/{heading_snakecase}.jac"\n'
            f'        ```\n    === "Python"\n'
            '        ```python linenums="1"\n'
            '        --8<-- "jac/examples/reference/'
            f'{heading_snakecase}.py"\n        ```\n'
            f'??? info "Jac Grammar Snippet"\n    ```yaml linenums="{lines[0]}"\n    --8<-- '
            f'"jac/jaclang/compiler/jac.lark:{lines[0]}:{lines[1]}"\n    ```\n'
            "**Description**\n\n--8<-- "
            f'"jac/examples/reference/'
            f'{heading_snakecase}.md"\n'
        )
        md_str += f"{content}\n"
    return md_str


def is_standard_lib_module(module_path: str) -> bool:
    """Check if a module is a standard library module."""
    import os
    import sysconfig

    stdlib_dir = sysconfig.get_paths()["stdlib"]
    direc_path = os.path.join(stdlib_dir, module_path)
    file_path = direc_path + ".py"
    return os.path.isfile(file_path) or os.path.isdir(direc_path)


def dump_traceback(e: Exception) -> str:
    """Dump the stack frames of the exception."""
    trace_dump = ""

    # Utility function to get the error line char offset.
    def byte_offset_to_char_offset(string: str, offset: int) -> int:
        return len(string.encode("utf-8")[:offset].decode("utf-8", errors="replace"))

    tb = TracebackException(type(e), e, e.__traceback__, limit=None, compact=True)
    trace_dump += f"Error: {str(e)}\n"

    # The first frame is the call the to the above `exec` function, not usefull to the enduser,
    # and Make the most recent call first.
    tb.stack.pop(0)
    tb.stack.reverse()

    # FIXME: should be some settings, we should replace to ensure the anchors length match.
    dump_tab_width = 2

    for idx, frame in enumerate(tb.stack):
        func_signature = frame.name + ("()" if frame.name.isidentifier() else "")

        # Pretty print the most recent call's location.
        if idx == 0 and (
            (frame.lineno is not None) and frame.line and frame.line.strip() != ""
        ):

            # Note: This is CPython internals we're trying to get since python doesn't provide
            # the frames original line but the stripped version so we had to do this.
            line_o = frame.line  # Fallback line.
            if hasattr(frame, "_original_line"):
                line_o = frame._original_line.rstrip()  # type: ignore [attr-defined]
            elif hasattr(frame, "_original_lines"):
                # https://github.com/python/cpython/issues/106922
                line_o = frame._original_lines.split("\n")[0].rstrip()  # type: ignore [attr-defined]

            if frame.colno is not None and frame.end_colno is not None:
                off_start = byte_offset_to_char_offset(line_o, frame.colno) - 1
                off_end = byte_offset_to_char_offset(line_o, frame.end_colno) - 1

                # Get the source.
                file_source = None
                with open(frame.filename, "r") as file:
                    file_source = file.read()

                # Get the source offset.
                lines = file_source.split("\n")
                for i in range(frame.lineno - 1):
                    off_start += len(lines[i]) + 1
                    off_end += len(lines[i]) + 1

                trace_dump += pretty_print_source_location(
                    frame.filename, file_source, frame.lineno, off_start, off_end
                )

        trace_dump += f'\n{" " * dump_tab_width}at {func_signature} {frame.filename}:{frame.lineno}'

    return trace_dump


# TODO: After implementing the TextRange (or simillar named) class to mark a text range
# refactor the parameter to accept an instace of that text range object.
def pretty_print_source_location(
    file_path: str,
    file_source: str,
    error_line: int,
    pos_start: int,
    pos_end: int,
) -> str:
    """Pretty print internal method for the pretty_print method."""
    # NOTE: The Line numbers and the column numbers are starts with 1.
    # We print totally 5 lines (error line and above 2 and bellow 2).

    # The width of the line number we'll be printing (more of a settings).
    line_num_width: int = 5

    idx: int = pos_start  # Pointer for the current character.

    if file_source == "" or file_path == "":
        return ""

    start_line: int = error_line - 2
    if start_line < 1:
        start_line = 1
    end_line: int = start_line + 5  # Index is exclusive.

    # Get the first character of the [start_line].
    file_source.splitlines(True)[start_line - 1]
    curr_line: int = error_line
    while idx >= 0 and curr_line >= start_line:
        idx -= 1
        if idx < 0:
            break
        if file_source[idx] == "\n":
            curr_line -= 1

    idx += 1  # Enter the line.
    assert idx == 0 or file_source[idx - 1] == "\n"

    pretty_dump = ""

    # Print each lines.
    curr_line = start_line
    while curr_line < end_line:
        pretty_dump += f"%{line_num_width}d | " % curr_line

        idx_line_start = idx
        while idx < len(file_source) and file_source[idx] != "\n":
            idx += 1  # Run to the line end.
        pretty_dump += file_source[idx_line_start:idx]
        pretty_dump += "\n"

        if curr_line == error_line:  # Print the current line with indicator.
            pretty_dump += f"%{line_num_width}s | " % " "

            spaces = ""
            for idx_pre in range(idx_line_start, pos_start):
                spaces += "\t" if file_source[idx_pre] == "\t" else " "

            err_token_len = pos_end - pos_start
            pretty_dump += spaces + ("^" * err_token_len) + "\n"

        if idx == len(file_source):
            break
        curr_line += 1
        idx += 1

    return pretty_dump[:-1]  # Get rid of the last newline (of the last line).


class Jdb(pdb.Pdb):
    """Jac debugger."""

    def __init__(self, *args, **kwargs) -> None:  # noqa
        """Initialize the Jac debugger."""
        super().__init__(*args, **kwargs)
        self.prompt = "Jdb > "

    def has_breakpoint(self, bytecode: bytes) -> bool:
        """Check for breakpoint."""
        code = marshal.loads(bytecode)
        instructions = dis.get_instructions(code)
        return any(
            instruction.opname in ("LOAD_GLOBAL", "LOAD_NAME", "LOAD_FAST")
            and instruction.argval == "breakpoint"
            for instruction in instructions
        )


debugger = Jdb()



"""Test Jac settings module."""

import os
from unittest.mock import mock_open, patch

from jaclang.settings import settings
from jaclang.utils.test import TestCase


class JacSettings(TestCase):
    """Test settings module."""

    def test_settings_config(self) -> None:
        """Basic settings for pass."""
        config_content = "[settings]\npass_timer = True"
        with patch("builtins.open", mock_open(read_data=config_content)):
            settings.load_config_file()
            self.assertEqual(settings.pass_timer, True)
        config_content = "[settings]\npass_timer = False"
        with patch("builtins.open", mock_open(read_data=config_content)):
            settings.load_config_file()
            self.assertEqual(settings.pass_timer, False)

    def test_settings_env_vars(self) -> None:
        """Basic settings for pass."""
        os.environ["JACLANG_PASS_TIMER"] = "True"
        settings.load_env_vars()
        self.assertEqual(settings.pass_timer, True)
        os.environ["JACLANG_PASS_TIMER"] = "False"
        settings.load_env_vars()
        self.assertEqual(settings.pass_timer, False)
        os.unsetenv("JACLANG_PASS_TIMER")

    def test_settings_precedence(self) -> None:
        """Basic settings for pass."""
        os.environ["JACLANG_PASS_TIMER"] = "True"
        config_content = "[settings]\npass_timer = False"
        with patch("builtins.open", mock_open(read_data=config_content)):
            settings.load_all()
            self.assertEqual(settings.pass_timer, True)
        config_content = "[settings]\npass_timer = True"
        os.environ["JACLANG_PASS_TIMER"] = "False"
        with patch("builtins.open", mock_open(read_data=config_content)):
            settings.load_all()
            self.assertEqual(settings.pass_timer, False)
        os.unsetenv("JACLANG_PASS_TIMER")



"""Control flow graph build pass for the Jac compiler.

This pass constructs a control flow graph (CFG) representation of the program by:
1. Identifying basic blocks (sequences of statements with a single entry and exit point)
2. Establishing control flow relationships between these blocks
3. Tracking loop structures and maintaining proper nesting relationships
4. Building connections between statements that affect control flow (conditionals, loops, etc.)

The CFG provides a foundation for data flow analysis, optimization, and understanding program execution paths.
The pass also includes functionality to coalesce basic blocks and generate visual representations of the CFG.
"""

from typing import Sequence

import jaclang.compiler.unitree as uni
from jaclang.compiler.passes import UniPass


class CFGBuildPass(UniPass):
    """Jac Symbol table build pass."""

    def before_pass(self) -> None:
        """Before pass."""
        self.while_loop_stack: list[list[uni.UniCFGNode]] = []
        self.for_loop_stack: list[list[uni.UniCFGNode]] = []
        self.ability_stack: list[list[uni.UniCFGNode]] = []
        self.to_connect: list[uni.UniCFGNode] = []
        self.first_exit: bool = False

    def push_loop_stack(self, loop_header: uni.UniCFGNode) -> None:
        """Push loop stack."""
        if isinstance(loop_header, uni.WhileStmt):
            self.while_loop_stack.append([loop_header])
        elif isinstance(loop_header, (uni.InForStmt, uni.IterForStmt)):
            self.for_loop_stack.append([loop_header])

    def pop_loop_stack(self, node: uni.UniCFGNode) -> None:
        """Pop loop stack."""
        if isinstance(node, uni.WhileStmt) and len(self.while_loop_stack) > 0:
            self.while_loop_stack.pop()
        elif (
            isinstance(node, (uni.InForStmt, uni.IterForStmt))
            and len(self.for_loop_stack) > 0
        ):
            self.for_loop_stack.pop()

    def get_parent_bb_stmt(self, node: uni.UniCFGNode) -> uni.UniCFGNode | None:
        """Get parent basic block."""
        if not isinstance(node, uni.Module):
            try:
                parent_bb_stmt = node.parent_of_type(uni.UniCFGNode)
                return parent_bb_stmt
            except Exception:
                return None
        else:
            return None

    def link_bbs(self, source: uni.UniCFGNode, target: uni.UniCFGNode) -> None:
        """Link basic blocks."""
        if source and target:
            if source.bb_out:
                source.bb_out.append(target)
            else:
                source.bb_out = [target]
            if target.bb_in:
                target.bb_in.append(source)
            else:
                target.bb_in = [source]

    def get_code_block_sequence(
        self, node: uni.CodeBlockStmt
    ) -> list[uni.UniCFGNode] | None:
        """Get code block sequence."""
        sequence: list[uni.UniCFGNode] = []
        if hasattr(node, "body") and isinstance(node.body, Sequence):
            for bbs in node.body:
                if isinstance(bbs, uni.UniCFGNode):
                    sequence.append(bbs)
            if sequence:
                return sequence
            else:
                return None
        else:
            return None

    def enter_node(self, node: uni.UniNode) -> None:
        """Enter BasicBlockStmt nodes."""
        if isinstance(node, uni.UniCFGNode) and not isinstance(node, uni.Semi):
            # check if the current node is a CodeBlockStmt in a sequence of statements
            bb_stmts = self.get_code_block_sequence(node.parent)
            if (
                isinstance(node, uni.CodeBlockStmt)
                and node.parent
                and (not (isinstance(node, (uni.ElseIf, uni.ElseStmt))))
                and bb_stmts
                and self.first_exit
            ):
                # bb_stmts = self.get_code_block_sequence(node.parent)
                # bb_stmts = [
                #     bbs for bbs in node.parent.body if isinstance(bbs, uni.UniCFGNode)
                # ]
                if (
                    node.parent
                    and isinstance(node.parent, uni.Archetype)
                    # and isinstance(node.parent.parent, uni.BasicBlockStmt)
                ):
                    parent_obj = node.parent
                    if parent_obj:
                        self.link_bbs(parent_obj, node)
                elif bb_stmts[0] == node:
                    if isinstance(node.parent, uni.ModuleCode) and self.to_connect:
                        for bb in self.to_connect:
                            self.link_bbs(bb, node)
                            self.to_connect.remove(bb)  # if self.to_connect:
                    else:
                        parent_bb = self.get_parent_bb_stmt(node)
                        if parent_bb:
                            self.link_bbs(parent_bb, node)
                elif self.to_connect and not isinstance(
                    node, (uni.ElseIf, uni.ElseStmt)
                ):
                    to_remove = []
                    for parent in self.to_connect:
                        if isinstance(parent, uni.UniCFGNode):
                            self.link_bbs(parent, node)
                            to_remove.append(parent)
                    for parent in to_remove:
                        self.to_connect.remove(parent)
                else:
                    for i in range(len(bb_stmts)):
                        if bb_stmts[i] == node:
                            parent_bb = bb_stmts[i - 1]
                            self.link_bbs(parent_bb, node)

            else:
                parent_bb = self.get_parent_bb_stmt(node)
                if parent_bb:
                    self.link_bbs(parent_bb, node)
            if isinstance(node, (uni.InForStmt, uni.IterForStmt, uni.WhileStmt)):
                self.push_loop_stack(node)
            else:
                if self.for_loop_stack:
                    self.for_loop_stack[-1].append(node)
                if self.while_loop_stack:
                    self.while_loop_stack[-1].append(node)
            if isinstance(node, uni.Ability):
                self.ability_stack.append([node])
            elif self.ability_stack:
                self.ability_stack[-1].append(node)

    def exit_node(self, node: uni.UniNode) -> None:
        """Exit BasicBlockStmt nodes."""
        if isinstance(node, uni.UniCFGNode) and not isinstance(node, uni.Semi):
            self.first_exit = True
            if not node.bb_out:
                self.to_connect.append(node)
            if (
                isinstance(node, (uni.InForStmt, uni.IterForStmt))
                and self.for_loop_stack
            ):
                for from_node in self.for_loop_stack[-1][1:]:
                    self.link_bbs(from_node, node)
                    if from_node in self.to_connect:
                        self.to_connect.remove(from_node)
                self.pop_loop_stack(node)
                self.to_connect.append(node)
            elif isinstance(node, uni.WhileStmt) and self.while_loop_stack:
                for from_node in self.while_loop_stack[-1][1:]:
                    self.link_bbs(from_node, node)
                    if from_node in self.to_connect:
                        self.to_connect.remove(from_node)
                self.pop_loop_stack(node)
                self.to_connect.append(node)
            elif isinstance(node, uni.Ability) and self.ability_stack:
                for from_node in self.ability_stack[-1][1:]:
                    if from_node in self.to_connect:
                        self.to_connect.remove(from_node)
                self.ability_stack.pop()

    def after_pass(self) -> None:
        """After pass."""
        if self.to_connect:
            for node in self.to_connect:
                if isinstance(node, uni.UniCFGNode):
                    parent_bb = self.get_parent_bb_stmt(node)
                    if parent_bb:
                        self.link_bbs(parent_bb, node)


class CoalesceBBPass(UniPass):
    """Fetch basic blocks."""

    def before_pass(self) -> None:
        """Before pass."""
        self.basic_blocks: dict = {}
        self.bb_counter: int = 0

    def get_bb(self, node: uni.UniCFGNode) -> list[uni.UniCFGNode]:
        """Get basic block."""
        head = node.get_head()
        tail = node.get_tail()
        bb_list: list[uni.UniCFGNode] = []
        if head and tail:
            cur_node = head
            while cur_node != tail:
                if cur_node not in bb_list:
                    bb_list.append(cur_node)
                if cur_node.bb_out:
                    cur_node = cur_node.bb_out[0]
            bb_list.append(tail) if tail not in bb_list else None

        return bb_list

    def enter_node(self, node: uni.UniNode) -> None:
        """Enter BasicBlockStmt nodes."""
        if isinstance(node, uni.UniCFGNode) and not isinstance(node, uni.Semi):
            tail = node.get_tail()
            bb = self.get_bb(node)
            for bbs in self.basic_blocks.values():
                if bbs["bb_stmts"] == bb:
                    return
            self.basic_blocks[self.bb_counter] = {
                "bb_stmts": bb,
                "bb_out": tail.bb_out,
            }
            self.bb_counter += 1

    def printgraph_cfg(self) -> str:
        """Generate dot graph for CFG."""
        cfg: dict = {}
        dot = "digraph G {\n"

        for key, bb in self.basic_blocks.items():
            cfg[key] = {
                "bb_stmts": bb["bb_stmts"],
                "bb_out": (
                    [0 for _ in range(len(bb["bb_out"]))] if bb["bb_out"] else None
                ),
            }

            if bb["bb_out"]:

                for out_obj in bb["bb_out"]:
                    for k, v in self.basic_blocks.items():
                        if out_obj == v["bb_stmts"][0]:
                            cfg[key]["bb_out"][bb["bb_out"].index(out_obj)] = k
                            break

        for key, bb in cfg.items():
            for i, bbs in enumerate(bb["bb_stmts"]):
                if isinstance(bbs, uni.ElseIf):
                    cfg[key]["bb_stmts"][i] = "elif " + bbs.condition.unparse()
                elif isinstance(bbs, uni.IfStmt):
                    cfg[key]["bb_stmts"][i] = "if " + bbs.condition.unparse()
                elif isinstance(bbs, uni.ElseStmt):
                    cfg[key]["bb_stmts"][i] = "else"
                elif isinstance(bbs, uni.Archetype):
                    cfg[key]["bb_stmts"][i] = "obj " + bbs.name.unparse()
                elif isinstance(bbs, uni.Ability):
                    cfg[key]["bb_stmts"][i] = (
                        "can " + bbs.name_ref.unparse() + bbs.signature.unparse()
                        if bbs.signature
                        else ""
                    )
                elif isinstance(bbs, (uni.InForStmt, uni.IterForStmt)):
                    cfg[key]["bb_stmts"][i] = (
                        (
                            "for "
                            + bbs.target.unparse()
                            + " in "
                            + bbs.collection.unparse()
                        )
                        if isinstance(bbs, uni.InForStmt)
                        else (
                            "for "
                            + bbs.iter.unparse()
                            + bbs.condition.unparse()
                            + bbs.count_by.unparse()
                        )
                    )
                elif isinstance(bbs, uni.WhileStmt):
                    cfg[key]["bb_stmts"][i] = "while " + bbs.condition.unparse()
                else:
                    cfg[key]["bb_stmts"][i] = bbs.unparse()

        for key, value in cfg.items():
            if value.get("bb_stmts"):
                stmts = "\\n".join(
                    [stmt.replace('"', '\\"') for stmt in value["bb_stmts"]]
                )
                dot += f'  {key} [label="BB{key}\\n{stmts}", shape=box];\n'
            else:
                dot += f'  {key} [label="BB{key}"];\n'

        for key, value in cfg.items():
            if value.get("bb_out"):
                for out_id in value["bb_out"]:
                    dot += f"  {key} -> {out_id};\n"

        dot += "}\n"
        return dot



"""Document Intermediate Representation (DocIR) classes."""

from __future__ import annotations

from typing import Optional, Union

# Define DocType for self-referential typing
DocType = Union["Doc", "Text", "Line", "Group", "Indent", "Concat", "IfBreak", "Align"]


class Doc:
    """Base class for document parts."""

    def __str__(self) -> str:
        """Return a string representation of the Doc object."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        """Return a string representation of the Doc object."""
        return self.treeprint()

    def treeprint(self, level: int = 0) -> str:
        """Return an indented string representation of the Doc object."""
        indent = "  " * level
        return f"{indent}{self.__class__.__name__}"


class Text(Doc):
    """Simple text content."""

    def __init__(self, text: str) -> None:
        """Initialize a Text object."""
        self.text: str = text

    def __str__(self) -> str:
        """Return a string representation of the Text object."""
        return f'Text("{self.text}")'

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        return f'{indent}Text("{self.text}")'


class Line(Doc):
    """Represents a line break that can be preserved or flattened."""

    def __init__(
        self, hard: bool = False, literal: bool = False, tight: bool = False
    ) -> None:
        """Initialize a Line object."""
        self.hard: bool = hard  # If True, always break
        self.literal: bool = literal  # includes the literal newline in flattened output
        self.tight: bool = tight  # If True, no space is added

    def __str__(self) -> str:
        """Return a string representation of the Line object."""
        attrs: list[str] = []
        if self.hard:
            attrs.append("hard")
        if self.literal:
            attrs.append("literal")
        return f"Line({', '.join(attrs)})"

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        attrs: list[str] = []
        if self.hard:
            attrs.append("hard=True")
        if self.literal:
            attrs.append("literal=True")
        return f"{indent}Line({', '.join(attrs)})"


class Group(Doc):
    """A group that can be printed flat or broken into multiple lines."""

    def __init__(
        self,
        contents: DocType,
        break_contiguous: bool = False,
        id: Optional[str] = None,
    ) -> None:
        """Initialize a Group object."""
        self.contents: DocType = contents
        self.break_contiguous: bool = break_contiguous
        self.id: Optional[str] = id

    def __str__(self) -> str:
        """Return a string representation of the Group object."""
        return f"Group({self.contents})"

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        header = (
            f"{indent}Group(id={self.id}, break_contiguous={self.break_contiguous}):"
        )
        children_repr: list[str] = [self.contents.treeprint(level + 1)]
        return f"{header}\n" + "\n".join(children_repr)


class Indent(Doc):
    """Indented content."""

    def __init__(self, contents: DocType) -> None:
        """Initialize an Indent object."""
        self.contents: DocType = contents

    def __str__(self) -> str:
        """Return a string representation of the Indent object."""
        return f"Indent({self.contents})"

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        header = f"{indent}Indent:"
        children_repr: list[str] = [self.contents.treeprint(level + 1)]
        return f"{header}\n" + "\n".join(children_repr)


class Concat(Doc):
    """A sequence of doc parts."""

    def __init__(self, parts: list[DocType]) -> None:
        """Initialize a Concat object."""
        self.parts: list[DocType] = parts

    def __str__(self) -> str:
        """Return a string representation of the Concat object."""
        return f"Concat({self.parts})"

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        header = f"{indent}Concat:"
        children_repr = [part.treeprint(level + 1) for part in self.parts]
        return f"{header}\n" + "\n".join(children_repr)


class IfBreak(Doc):
    """Content that differs based on whether the parent group is broken."""

    def __init__(self, break_contents: DocType, flat_contents: DocType) -> None:
        """Initialize an IfBreak object."""
        self.break_contents: DocType = break_contents
        self.flat_contents: DocType = flat_contents

    def __str__(self) -> str:
        """Return a string representation of the IfBreak object."""
        return f"IfBreak({self.break_contents}, {self.flat_contents})"

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        header = f"{indent}IfBreak:"
        break_repr: list[str] = [f"{indent}  break_contents:"]
        break_repr.append(self.break_contents.treeprint(level + 2))

        flat_repr: list[str] = [f"{indent}  flat_contents:"]
        flat_repr.append(self.flat_contents.treeprint(level + 2))

        return f"{header}\n" + "\n".join(break_repr) + "\n" + "\n".join(flat_repr)


class Align(Doc):
    """Alignment relative to the current indentation level."""

    def __init__(self, contents: DocType, n: Optional[int] = None) -> None:
        """Initialize an Align object."""
        self.contents: DocType = contents
        self.n: Optional[int] = (
            n  # Number of spaces, or null to use current indentation level
        )

    def __str__(self) -> str:
        """Return a string representation of the Align object."""
        return f"Align({self.n}, {self.contents})"

    def treeprint(self, level: int = 0) -> str:
        indent = "  " * level
        header = f"{indent}Align(n={self.n}):"
        children_repr: list[str] = [self.contents.treeprint(level + 1)]
        return f"{header}\n" + "\n".join(children_repr)



"""Standardized transformation process and error interface."""

from __future__ import annotations

import time
from abc import ABC, abstractmethod
from typing import Generic, Optional, TYPE_CHECKING, Type, TypeVar

from jaclang.compiler.codeinfo import CodeLocInfo
from jaclang.compiler.unitree import UniNode
from jaclang.settings import settings
from jaclang.utils.helpers import pretty_print_source_location
from jaclang.utils.log import logging

if TYPE_CHECKING:
    from jaclang.compiler.program import JacProgram

T = TypeVar("T", bound=UniNode)
R = TypeVar("R", bound=UniNode)


class Alert:
    """Alert interface."""

    def __init__(self, msg: str, loc: CodeLocInfo, from_pass: Type[Transform]) -> None:
        """Initialize alert."""
        self.msg = msg
        self.loc: CodeLocInfo = loc
        self.from_pass: Type[Transform] = from_pass

    def __str__(self) -> str:
        """Return string representation of alert."""
        return (
            f" {self.loc.mod_path}, line {self.loc.first_line},"
            f" col {self.loc.col_start}: {self.msg}"
        )

    def __repr__(self) -> str:
        """Return string representation of alert."""
        return self.as_log()

    def as_log(self) -> str:
        """Return the alert as a single line log as opposed to the pretty print."""
        file_path: str = self.loc.mod_path
        if file_path == "":
            return self.msg  # There are error messages without file references.

        line: int = self.loc.first_line
        column: int = self.loc.col_start
        return f"{file_path}:{line}:{column} {self.msg}"

    def pretty_print(self) -> str:
        """Pretty pritns the Alert to show the alert with source location."""
        pretty_dump = pretty_print_source_location(
            self.loc.mod_path,
            self.loc.orig_src.code,
            self.loc.first_line,
            self.loc.pos_start,
            self.loc.pos_end,
        )
        if pretty_dump != "":
            pretty_dump = "\n" + pretty_dump
        return self.as_log() + pretty_dump


class Transform(ABC, Generic[T, R]):
    """Abstract class for IR passes."""

    def __init__(self, ir_in: T, prog: JacProgram) -> None:
        """Initialize pass."""
        self.logger = logging.getLogger(self.__class__.__name__)
        self.errors_had: list[Alert] = []
        self.warnings_had: list[Alert] = []
        self.cur_node: UniNode = ir_in  # tracks current node during traversal
        self.prog = prog
        self.time_taken = 0.0
        self.ir_in: T = ir_in
        self.pre_transform()
        self.ir_out: R = self.timed_transform(ir_in=ir_in)
        self.post_transform()

    def timed_transform(
        self,
        ir_in: T,
    ) -> R:
        """Transform with time tracking."""
        start_time = time.time()
        ir_out = self.transform(ir_in=ir_in)
        self.time_taken = time.time() - start_time
        if settings.pass_timer:
            self.log_info(
                f"Time taken in {self.__class__.__name__}: {self.time_taken:.4f} seconds"
            )
        return ir_out

    def pre_transform(self) -> None:
        """Pre-transform hook."""
        pass

    def post_transform(self) -> None:
        """Post-transform hook."""
        pass

    @abstractmethod
    def transform(self, ir_in: T) -> R:
        """Transform interface."""
        pass

    def log_error(self, msg: str, node_override: Optional[UniNode] = None) -> None:
        """Pass Error."""
        alrt = Alert(
            msg,
            self.cur_node.loc if not node_override else node_override.loc,
            self.__class__,
        )
        self.errors_had.append(alrt)
        self.prog.errors_had.append(alrt)
        self.logger.error(alrt.as_log())

    def log_warning(self, msg: str, node_override: Optional[UniNode] = None) -> None:
        """Pass Error."""
        alrt = Alert(
            msg,
            self.cur_node.loc if not node_override else node_override.loc,
            self.__class__,
        )
        self.warnings_had.append(alrt)
        self.prog.warnings_had.append(alrt)
        self.logger.warning(alrt.as_log())

    def log_info(self, msg: str) -> None:
        """Log info."""
        self.logger.info(msg)

    def ice(self, msg: str = "Something went horribly wrong!") -> RuntimeError:
        """Pass Error."""
        self.log_error(f"ICE: Pass {self.__class__.__name__} - {msg}")
        return RuntimeError(
            f"Internal Compiler Error: Pass {self.__class__.__name__} - {msg}"
        )



from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["voterank"]

@_dispatchable
def voterank(G: Graph[_Node], number_of_nodes: int | None = None): ...



"""Symbol Table Construction Pass for the Jac compiler.

This pass builds the hierarchical symbol table structure for the entire program by:

1. Creating symbol tables for each scope in the program (modules, archetypes, abilities, blocks)
2. Establishing parent-child relationships between nested scopes
3. Registering symbols for various language constructs:
   - Global variables and imports
   - Archetypes (objects, nodes, edges, walkers) and their members
   - Abilities (methods and functions) and their parameters
   - Enums and their values
   - Local variables in various block scopes

4. Adding special symbols like 'self' and 'super' in appropriate contexts
5. Maintaining scope boundaries for proper symbol resolution

The symbol table is a fundamental data structure that enables name resolution,
type checking, and semantic analysis throughout the compilation process.
"""

import jaclang.compiler.unitree as uni
from jaclang.compiler.passes import UniPass
from jaclang.compiler.unitree import UniScopeNode


class SymTabBuildPass(UniPass):
    """Jac Symbol table build pass."""

    def before_pass(self) -> None:
        """Before pass."""
        self.cur_sym_tab: list[UniScopeNode] = []

    def push_scope_and_link(self, key_node: uni.UniScopeNode) -> None:
        """Push scope."""
        if not len(self.cur_sym_tab):
            self.cur_sym_tab.append(key_node)
        else:
            self.cur_sym_tab.append(self.cur_scope.link_kid_scope(key_node=key_node))

    def pop_scope(self) -> UniScopeNode:
        """Pop scope."""
        return self.cur_sym_tab.pop()

    @property
    def cur_scope(self) -> UniScopeNode:
        """Return current scope."""
        return self.cur_sym_tab[-1]

    def enter_module(self, node: uni.Module) -> None:
        self.push_scope_and_link(node)

    def exit_module(self, node: uni.Module) -> None:
        self.pop_scope()

    def exit_global_vars(self, node: uni.GlobalVars) -> None:
        for i in self.get_all_sub_nodes(node, uni.Assignment):
            for j in i.target:
                if isinstance(j, uni.AstSymbolNode):
                    j.sym_tab.def_insert(j, access_spec=node, single_decl="global var")
                else:
                    self.ice("Expected name type for globabl vars")

    def enter_test(self, node: uni.Test) -> None:
        self.push_scope_and_link(node)
        import unittest

        for i in [j for j in dir(unittest.TestCase()) if j.startswith("assert")]:
            node.sym_tab.def_insert(
                uni.Name.gen_stub_from_node(node, i, set_name_of=node)
            )

    def exit_test(self, node: uni.Test) -> None:
        self.pop_scope()

    def exit_module_path(self, node: uni.ModulePath) -> None:
        if node.alias:
            node.alias.sym_tab.def_insert(node.alias, single_decl="import")
        elif node.path and isinstance(node.path[0], uni.Name):
            if node.parent_of_type(uni.Import) and not (
                node.parent_of_type(uni.Import).from_loc
                and node.parent_of_type(uni.Import).is_jac
            ):
                node.path[0].sym_tab.def_insert(node.path[0])
        else:
            pass  # Need to support pythonic import symbols with dots in it

    def enter_archetype(self, node: uni.Archetype) -> None:
        self.push_scope_and_link(node)
        assert node.parent_scope is not None
        node.parent_scope.def_insert(node, access_spec=node, single_decl="archetype")

    def exit_archetype(self, node: uni.Archetype) -> None:
        self.pop_scope()

    def enter_ability(self, node: uni.Ability) -> None:
        self.push_scope_and_link(node)
        assert node.parent_scope is not None
        node.parent_scope.def_insert(node, access_spec=node, single_decl="ability")
        if node.is_method:
            node.sym_tab.def_insert(uni.Name.gen_stub_from_node(node, "self"))
            node.sym_tab.def_insert(
                uni.Name.gen_stub_from_node(
                    node, "super", set_name_of=node.method_owner
                )
            )

    def exit_ability(self, node: uni.Ability) -> None:
        self.pop_scope()

    def enter_impl_def(self, node: uni.ImplDef) -> None:
        self.push_scope_and_link(node)
        assert node.parent_scope is not None
        node.parent_scope.def_insert(node, single_decl="impl")

    def exit_impl_def(self, node: uni.ImplDef) -> None:
        self.pop_scope()

    def enter_sem_def(self, node: uni.SemDef) -> None:
        self.push_scope_and_link(node)
        assert node.parent_scope is not None
        node.parent_scope.def_insert(node, single_decl="sem")

    def exit_sem_def(self, node: uni.SemDef) -> None:
        self.pop_scope()

    def enter_enum(self, node: uni.Enum) -> None:
        self.push_scope_and_link(node)
        assert node.parent_scope is not None
        node.parent_scope.def_insert(node, access_spec=node, single_decl="enum")

    def exit_enum(self, node: uni.Enum) -> None:
        self.pop_scope()

    def enter_typed_ctx_block(self, node: uni.TypedCtxBlock) -> None:
        self.push_scope_and_link(node)

    def exit_typed_ctx_block(self, node: uni.TypedCtxBlock) -> None:
        self.pop_scope()

    def enter_if_stmt(self, node: uni.IfStmt) -> None:
        self.push_scope_and_link(node)

    def exit_if_stmt(self, node: uni.IfStmt) -> None:
        self.pop_scope()

    def enter_else_if(self, node: uni.ElseIf) -> None:
        self.push_scope_and_link(node)

    def exit_else_if(self, node: uni.ElseIf) -> None:
        self.pop_scope()

    def enter_else_stmt(self, node: uni.ElseStmt) -> None:
        self.push_scope_and_link(node)

    def exit_else_stmt(self, node: uni.ElseStmt) -> None:
        self.pop_scope()

    def enter_try_stmt(self, node: uni.TryStmt) -> None:
        self.push_scope_and_link(node)

    def exit_try_stmt(self, node: uni.TryStmt) -> None:
        self.pop_scope()

    def enter_except(self, node: uni.Except) -> None:
        self.push_scope_and_link(node)

    def exit_except(self, node: uni.Except) -> None:
        self.pop_scope()

    def enter_finally_stmt(self, node: uni.FinallyStmt) -> None:
        self.push_scope_and_link(node)

    def exit_finally_stmt(self, node: uni.FinallyStmt) -> None:
        self.pop_scope()

    def enter_iter_for_stmt(self, node: uni.IterForStmt) -> None:
        self.push_scope_and_link(node)

    def exit_iter_for_stmt(self, node: uni.IterForStmt) -> None:
        self.pop_scope()

    def enter_in_for_stmt(self, node: uni.InForStmt) -> None:
        self.push_scope_and_link(node)

    def exit_in_for_stmt(self, node: uni.InForStmt) -> None:
        self.pop_scope()

    def enter_while_stmt(self, node: uni.WhileStmt) -> None:
        self.push_scope_and_link(node)

    def exit_while_stmt(self, node: uni.WhileStmt) -> None:
        self.pop_scope()

    def enter_with_stmt(self, node: uni.WithStmt) -> None:
        self.push_scope_and_link(node)

    def exit_with_stmt(self, node: uni.WithStmt) -> None:
        self.pop_scope()

    def enter_lambda_expr(self, node: uni.LambdaExpr) -> None:
        self.push_scope_and_link(node)

    def exit_lambda_expr(self, node: uni.LambdaExpr) -> None:
        self.pop_scope()

    def enter_inner_compr(self, node: uni.InnerCompr) -> None:
        self.push_scope_and_link(node)

    def exit_inner_compr(self, node: uni.InnerCompr) -> None:
        self.pop_scope()

    def enter_dict_compr(self, node: uni.DictCompr) -> None:
        self.push_scope_and_link(node)

    def exit_dict_compr(self, node: uni.DictCompr) -> None:
        self.pop_scope()

    def enter_match_case(self, node: uni.MatchCase) -> None:
        self.push_scope_and_link(node)

    def exit_match_case(self, node: uni.MatchCase) -> None:
        self.pop_scope()



"""Test pass module."""

import io
import sys

import jaclang.compiler.unitree as uni
from jaclang import JacMachine as Jac
from jaclang.cli import cli
from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCase


class DeclImplMatchPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        Jac.set_base_path(self.fixture_abs_path("./"))
        Jac.attach_program(
            JacProgram(),
        )
        return super().setUp()

    def test_parameter_count_mismatch(self) -> None:
        """Basic test for pass."""
        (out := JacProgram()).compile(self.fixture_abs_path("defn_decl_mismatch.jac"))

        expected_stdout_values = (
            "Parameter count mismatch for ability impl.SomeObj.foo.",
            "    8 |",
            "    9 | # Miss match parameter count.",
            "   10 | impl SomeObj.foo(param1: str) -> str {",
            "      |      ^^^^^^^^^^^",
            '   11 |     return "foo";',
            "   12 | }",
            "From the declaration of foo.",
            "    2 |",
            "    3 | obj SomeObj {",
            "    4 |     def foo(param1: str, param2:int) -> str;",
            "      |         ^^^",
            "    5 |     def bar(param1: str, param2:int) -> str;",
            "    6 | }",
        )

        errors_output = ""
        for error in out.errors_had:
            errors_output += error.pretty_print() + "\n"

        print(errors_output)
        for exp in expected_stdout_values:
            self.assertIn(exp, errors_output)

    def test_ability_connected_to_decl(self) -> None:
        """Basic test for pass."""
        state = (out := JacProgram()).compile(self.fixture_abs_path("base.jac"))
        self.assertFalse(out.errors_had)
        self.assertIn("impl.Test.say_hi", state.impl_mod[0].sym_tab.names_in_scope)
        self.assertIsNotNone(
            state.impl_mod[0]
            .sym_tab.names_in_scope["impl.Test.say_hi"]
            .decl.name_of.body
        )
        self.assertIn("impl.Test.__init__", state.impl_mod[0].sym_tab.names_in_scope)
        self.assertIsNotNone(
            state.impl_mod[0]
            .sym_tab.names_in_scope["impl.Test.__init__"]
            .decl.name_of.body
        )

    def test_ability_connected_to_decl_post(self) -> None:
        """Basic test for pass."""
        state = (out := JacProgram()).compile(self.fixture_abs_path("base2.jac"))
        self.assertFalse(out.errors_had)
        self.assertIn("impl.Test.say_hi", state.sym_tab.impl_mod[0].names_in_scope)
        self.assertIsNotNone(
            state.sym_tab.impl_mod[0]
            .names_in_scope["impl.Test.say_hi"]
            .decl.name_of.body
        )
        self.assertIn("impl.Test.__init__", state.impl_mod[0].sym_tab.names_in_scope)
        self.assertIsNotNone(
            state.impl_mod[0]
            .sym_tab.names_in_scope["impl.Test.__init__"]
            .decl.name_of.body
        )

    def test_run_base2(self) -> None:
        """Test that the walker and node can be created dynamically."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.run(self.fixture_abs_path("base2.jac"))
        output = captured_output.getvalue().strip()
        self.assertIn("56", output)

    def test_arch_ref_has_sym(self) -> None:
        """Basic test for pass."""
        state = JacProgram().compile(self.fixture_abs_path("defs_and_uses.jac"))
        for i in state.get_all_sub_nodes(uni.ImplDef):
            self.assertIsNotNone(i.sym)

    def test_single_impl_annex(self) -> None:
        """Basic test for pass."""
        mypass = JacProgram().compile(
            self.examples_abs_path("manual_code/circle_pure.jac")
        )
        self.assertEqual(mypass.impl_mod[0].pp().count("ImplDef - impl.Circle.area"), 1)

    def test_impl_decl_resolution_fix(self) -> None:
        """Test walking through edges and nodes."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import("mtest", base_path=self.fixture_abs_path("./"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("2.0\n", stdout_value)

    def test_impl_grab(self) -> None:
        """Test walking through edges."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import("impl_grab", base_path=self.fixture_abs_path("./"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("1.414", stdout_value)

    def test_nested_impls(self) -> None:
        """Test complex nested impls."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import("nested_impls", base_path=self.fixture_abs_path("./"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue().split("\n")
        self.assertIn("Hello,from bar in kk", stdout_value[0])
        self.assertIn("Greeting: Hello, World!", stdout_value[1])
        self.assertIn("Repeated: Hello", stdout_value[2])
        self.assertIn("Hello, World!", stdout_value[3])
        self.assertIn("Last message:!", stdout_value[4])
        self.assertIn("Final message:!", stdout_value[5])

    def test_abstraction_bug(self) -> None:
        """Parse micro jac file."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import("atest", base_path=self.fixture_abs_path("./"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertEqual(stdout_value, "42\n")

    def test_inner_mod_impl(self) -> None:
        """Parse micro jac file."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        Jac.jac_import("enumerations", base_path=self.fixture_abs_path("./"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertEqual(stdout_value, "1\n")



"""JacFormatPass for Jaseci Ast.

This is a pass for formatting Jac code.
"""

from typing import Optional

import jaclang.compiler.passes.tool.doc_ir as doc
import jaclang.compiler.unitree as uni
from jaclang.compiler.passes import Transform
from jaclang.settings import settings


class JacFormatPass(Transform[uni.Module, uni.Module]):
    """JacFormat Pass format Jac code."""

    def pre_transform(self) -> None:
        """Initialize pass."""
        self.indent_size = 4
        self.MAX_LINE_LENGTH = settings.max_line_length

    def transform(self, ir_in: uni.Module) -> uni.Module:
        """After pass."""
        ir_in.gen.jac = self.format_doc_ir()
        return ir_in

    def format_doc_ir(
        self,
        doc_node: Optional[doc.DocType] = None,
        indent_level: int = 0,
        width_remaining: Optional[int] = None,
        is_broken: bool = False,
    ) -> str:
        """Recursively print a Doc node or a list of Doc nodes."""
        if doc_node is None:
            doc_node = self.ir_in.gen.doc_ir

        if width_remaining is None:
            width_remaining = self.MAX_LINE_LENGTH

        if isinstance(doc_node, doc.Text):
            return doc_node.text

        elif isinstance(doc_node, doc.Line):
            if is_broken or doc_node.hard:
                return "\n" + " " * (indent_level * self.indent_size)
            elif doc_node.literal:  # literal soft line
                return "\n"
            elif doc_node.tight:
                return ""
            else:  # soft line, not broken
                return " "

        elif isinstance(doc_node, doc.Group):
            # Try to print flat first. For this attempt, the group itself isn't forced to break.
            flat_contents_str = self.format_doc_ir(
                doc_node.contents, indent_level, width_remaining, is_broken=False
            )
            if (
                "\n" not in flat_contents_str
                and len(flat_contents_str) <= width_remaining
            ):
                return flat_contents_str
            else:
                full_width_for_broken_content = self.MAX_LINE_LENGTH - (
                    indent_level * self.indent_size
                )
                return self.format_doc_ir(
                    doc_node.contents,
                    indent_level,
                    full_width_for_broken_content,
                    is_broken=True,
                )

        elif isinstance(doc_node, doc.Indent):
            new_indent_level = indent_level + 1

            width_for_indented_content = self.MAX_LINE_LENGTH - (
                new_indent_level * self.indent_size
            )
            return self.format_doc_ir(
                doc_node.contents,
                new_indent_level,
                width_for_indented_content,  # Budget for lines within indent
                is_broken,  # is_broken state propagates
            )

        elif isinstance(doc_node, doc.Concat):
            result = ""
            # current_line_budget is the space left on the current line for the current part.
            current_line_budget = width_remaining

            for part in doc_node.parts:
                part_str = self.format_doc_ir(
                    part, indent_level, current_line_budget, is_broken
                )
                if part_str.startswith("\n"):
                    result = result.rstrip(" ")
                result += part_str

                if "\n" in part_str:
                    # part_str created a newline. The next part starts on a new line.
                    # Its budget is the full width available at this indent level.
                    current_line_budget = self.MAX_LINE_LENGTH - (
                        indent_level * self.indent_size
                    )
                    # Subtract what the *last line* of part_str consumed from this budget.
                    # The characters on the last line after the indent string.
                    indent_str_len = indent_level * self.indent_size
                    last_line_of_part = part_str.splitlines()[-1]

                    content_on_last_line = 0
                    if last_line_of_part.startswith(" " * indent_str_len):
                        content_on_last_line = len(last_line_of_part) - indent_str_len
                    else:  # It was a line not starting with the full indent (e.g. literal \n)
                        content_on_last_line = len(last_line_of_part)

                    current_line_budget -= content_on_last_line
                else:
                    # part_str stayed on the same line. Reduce budget for next part on this line.
                    current_line_budget -= len(part_str)

                if current_line_budget < 0:  # Ensure budget isn't negative
                    current_line_budget = 0
            return result

        elif isinstance(doc_node, doc.IfBreak):
            if is_broken:
                return self.format_doc_ir(
                    doc_node.break_contents, indent_level, width_remaining, is_broken
                )
            else:
                return self.format_doc_ir(
                    doc_node.flat_contents, indent_level, width_remaining, is_broken
                )

        elif isinstance(doc_node, doc.Align):
            align_spaces = doc_node.n if doc_node.n is not None else self.indent_size
            # effective_total_indent_spaces_for_children = (
            #     indent_level * self.indent_size
            # ) + align_spaces
            child_indent_level_for_align = indent_level + (
                align_spaces // self.indent_size
            )

            child_width_budget = width_remaining - align_spaces
            if child_width_budget < 0:
                child_width_budget = 0

            return self.format_doc_ir(
                doc_node.contents,
                child_indent_level_for_align,  # Approximated level for Lines inside
                child_width_budget,  # Budget for content on first line
                is_broken,
            )
        else:
            raise ValueError(f"Unknown DocType: {type(doc_node)}")



"""Jac Machine module."""

from __future__ import annotations

import ast as py_ast
import marshal
import types
from typing import Optional

import jaclang.compiler.unitree as uni
from jaclang.compiler.parser import JacParser
from jaclang.compiler.passes.main import (
    Alert,
    CFGBuildPass,
    DeclImplMatchPass,
    DefUsePass,
    InheritancePass,
    JacAnnexPass,
    JacImportDepsPass,
    PyBytecodeGenPass,
    PyJacAstLinkPass,
    PyastBuildPass,
    PyastGenPass,
    SemDefMatchPass,
    SymTabBuildPass,
    SymTabLinkPass,
    Transform,
)
from jaclang.compiler.passes.tool import (
    DocIRGenPass,
    FuseCommentsPass,
    JacFormatPass,
)
from jaclang.utils.log import logging


logger = logging.getLogger(__name__)

ir_gen_sched = [
    SymTabBuildPass,
    DeclImplMatchPass,
    DefUsePass,
    SemDefMatchPass,
    CFGBuildPass,
    InheritancePass,
]
py_code_gen = [
    PyastGenPass,
    PyJacAstLinkPass,
    PyBytecodeGenPass,
]
format_sched = [FuseCommentsPass, DocIRGenPass, JacFormatPass]


class JacProgram:
    """JacProgram to handle the Jac program-related functionalities."""

    def __init__(self, main_mod: Optional[uni.ProgramModule] = None) -> None:
        """Initialize the JacProgram object."""
        self.mod: uni.ProgramModule = main_mod if main_mod else uni.ProgramModule()
        self.py_raise_map: dict[str, str] = {}
        self.errors_had: list[Alert] = []
        self.warnings_had: list[Alert] = []

    def get_bytecode(self, full_target: str) -> Optional[types.CodeType]:
        """Get the bytecode for a specific module."""
        if full_target in self.mod.hub and self.mod.hub[full_target].gen.py_bytecode:
            codeobj = self.mod.hub[full_target].gen.py_bytecode
            return marshal.loads(codeobj) if isinstance(codeobj, bytes) else None
        result = self.compile(file_path=full_target)
        return marshal.loads(result.gen.py_bytecode) if result.gen.py_bytecode else None

    def parse_str(self, source_str: str, file_path: str) -> uni.Module:
        """Convert a Jac file to an AST."""
        had_error = False
        if file_path.endswith(".py"):
            parsed_ast = py_ast.parse(source_str)
            py_ast_ret = PyastBuildPass(
                ir_in=uni.PythonModuleAst(
                    parsed_ast,
                    orig_src=uni.Source(source_str, mod_path=file_path),
                ),
                prog=self,
            )
            had_error = len(py_ast_ret.errors_had) > 0
            mod = py_ast_ret.ir_out
        else:
            source = uni.Source(source_str, mod_path=file_path)
            jac_ast_ret: Transform[uni.Source, uni.Module] = JacParser(
                root_ir=source, prog=self
            )
            had_error = len(jac_ast_ret.errors_had) > 0
            mod = jac_ast_ret.ir_out
        if had_error:
            return mod
        if self.mod.main.stub_only:
            self.mod = uni.ProgramModule(mod)
        self.mod.hub[mod.loc.mod_path] = mod
        JacAnnexPass(ir_in=mod, prog=self)
        return mod

    def compile(
        self, file_path: str, use_str: str | None = None, no_cgen: bool = False
    ) -> uni.Module:
        """Convert a Jac file to an AST."""
        if not use_str:
            with open(file_path, "r", encoding="utf-8") as file:
                use_str = file.read()
        mod_targ = self.parse_str(use_str, file_path)
        self.run_schedule(mod=mod_targ, passes=ir_gen_sched)
        if not no_cgen:
            self.run_schedule(mod=mod_targ, passes=py_code_gen)
        return mod_targ

    def build(self, file_path: str, use_str: str | None = None) -> uni.Module:
        """Convert a Jac file to an AST."""
        mod_targ = self.compile(file_path, use_str)
        JacImportDepsPass(ir_in=mod_targ, prog=self)
        for mod in self.mod.hub.values():
            SymTabLinkPass(ir_in=mod, prog=self)
        for mod in self.mod.hub.values():
            DefUsePass(mod, prog=self)
        return mod_targ

    def run_schedule(
        self,
        mod: uni.Module,
        passes: list[type[Transform[uni.Module, uni.Module]]],
    ) -> None:
        """Run the passes on the module."""
        final_pass: Optional[type[Transform[uni.Module, uni.Module]]] = None
        for current_pass in passes:
            if current_pass == PyBytecodeGenPass:
                final_pass = current_pass
                break
            current_pass(ir_in=mod, prog=self)  # type: ignore
        if final_pass:
            final_pass(mod, prog=self)

    @staticmethod
    def jac_file_formatter(file_path: str) -> str:
        """Convert a Jac file to an AST."""
        prog = JacProgram()
        with open(file_path) as file:
            source = uni.Source(file.read(), mod_path=file_path)
            prse: Transform = JacParser(root_ir=source, prog=prog)
        for i in format_sched:
            prse = i(ir_in=prse.ir_out, prog=prog)
        prse.errors_had = prog.errors_had
        prse.warnings_had = prog.warnings_had
        return prse.ir_out.gen.jac

    @staticmethod
    def jac_str_formatter(source_str: str, file_path: str) -> str:
        """Convert a Jac file to an AST."""
        prog = JacProgram()
        source = uni.Source(source_str, mod_path=file_path)
        prse: Transform = JacParser(root_ir=source, prog=prog)
        for i in format_sched:
            prse = i(ir_in=prse.ir_out, prog=prog)
        prse.errors_had = prog.errors_had
        prse.warnings_had = prog.warnings_had
        return prse.ir_out.gen.jac if not prse.errors_had else source_str



"""Command line interface tool for the Jac language."""

import ast as ast3
import importlib
import marshal
import os
import pickle
import sys
import types
from pathlib import Path
from typing import Optional

import jaclang.compiler.unitree as uni
from jaclang.cli.cmdreg import CommandShell, cmd_registry
from jaclang.compiler.passes.main import PyastBuildPass
from jaclang.compiler.program import JacProgram
from jaclang.runtimelib.builtin import printgraph
from jaclang.runtimelib.constructs import WalkerArchetype
from jaclang.runtimelib.machine import (
    ExecutionContext,
    JacMachine as Jac,
    JacMachineInterface as JacInterface,
)
from jaclang.utils.helpers import debugger as db
from jaclang.utils.lang_tools import AstTool


JacInterface.create_cmd()
Jac.setup()


@cmd_registry.register
def format(path: str, outfile: str = "", to_screen: bool = False) -> None:
    """Format .jac files with improved code style.

    Applies consistent formatting to Jac code files to improve readability and
    maintain a standardized code style across your project.

    Args:
        path: Path to a .jac file or directory containing .jac files
        outfile: Optional output file path (when formatting a single file)
        to_screen: Print formatted code to stdout instead of writing to file

    Examples:
        jac format myfile.jac
        jac format myproject/
        jac format myfile.jac --outfile formatted.jac
        jac format myfile.jac --to_screen
    """

    def write_formatted_code(code: str, target_path: str) -> None:
        """Write formatted code to the appropriate destination."""
        if to_screen:
            print(code)
        elif outfile:
            with open(outfile, "w") as f:
                f.write(code)
        else:
            with open(target_path, "w") as f:
                f.write(code)

    path_obj = Path(path)

    # Case 1: Single .jac file
    if path.endswith(".jac"):
        if not path_obj.exists():
            print(f"Error: File '{path}' does not exist.", file=sys.stderr)
            return
        formatted_code = JacProgram.jac_file_formatter(str(path_obj))
        write_formatted_code(formatted_code, str(path_obj))
        return

    # Case 2: Directory with .jac files
    if path_obj.is_dir():
        jac_files = list(path_obj.glob("**/*.jac"))
        for jac_file in jac_files:
            formatted_code = JacProgram.jac_file_formatter(str(jac_file))
            write_formatted_code(formatted_code, str(jac_file))

        print(f"Formatted {len(jac_files)} '.jac' files.", file=sys.stderr)
        return

    # Case 3: Invalid path
    print(f"Error: '{path}' is not a .jac file or directory.", file=sys.stderr)


def proc_file_sess(
    filename: str, session: str, root: Optional[str] = None
) -> tuple[str, str, ExecutionContext]:
    """Create JacMachine and return the base path, module name, and machine state."""
    if session == "":
        session = (
            cmd_registry.args.session
            if hasattr(cmd_registry, "args")
            and hasattr(cmd_registry.args, "session")
            and cmd_registry.args.session
            else ""
        )
    base, mod = os.path.split(filename)
    base = base if base else "./"
    mod = mod[:-4]
    mach = ExecutionContext(session=session, root=root)
    Jac.set_context(mach)
    return base, mod, mach


@cmd_registry.register
def run(
    filename: str,
    session: str = "",
    main: bool = True,
    cache: bool = True,
) -> None:
    """Run the specified .jac file.

    Executes a Jac program file, loading it into the Jac runtime environment
    and running its code. This is the primary way to execute Jac programs.

    Args:
        filename: Path to the .jac or .jir file to run
        session: Optional session identifier for persistent state
        main: Treat the module as __main__ (default: True)
        cache: Use cached compilation if available (default: True)

    Examples:
        jac run myprogram.jac
        jac run myprogram.jac --session mysession
        jac run myprogram.jac --no-main
    """
    # if no session specified, check if it was defined when starting the command shell
    # otherwise default to jaclang.session
    base, mod, mach = proc_file_sess(filename, session)
    Jac.set_base_path(base)

    if filename.endswith(".jac"):
        try:
            Jac.jac_import(
                target=mod,
                base_path=base,
                override_name="__main__" if main else None,
            )
        except Exception as e:
            print(e, file=sys.stderr)
    elif filename.endswith(".jir"):
        try:
            with open(filename, "rb") as f:
                Jac.attach_program(pickle.load(f))
                Jac.jac_import(
                    target=mod,
                    base_path=base,
                    override_name="__main__" if main else None,
                )
        except Exception as e:
            print(e, file=sys.stderr)

    else:
        print("Not a valid file!\nOnly supports `.jac` and `.jir`")
    mach.close()


@cmd_registry.register
def get_object(filename: str, id: str, session: str = "", main: bool = True) -> dict:
    """Get the object with the specified id.

    Retrieves a specific object from a Jac program by its unique identifier.
    Returns the object's state as a dictionary.

    Args:
        filename: Path to the .jac or .jir file containing the object
        id: Unique identifier of the object to retrieve
        session: Optional session identifier for persistent state
        main: Treat the module as __main__ (default: True)

    Examples:
        jac get_object myprogram.jac obj123
        jac get_object myprogram.jac obj123 --session mysession
    """
    base, mod, mach = proc_file_sess(filename, session)

    if filename.endswith(".jac"):
        Jac.jac_import(
            target=mod,
            base_path=base,
            override_name="__main__" if main else None,
        )
    elif filename.endswith(".jir"):
        with open(filename, "rb") as f:
            Jac.attach_program(pickle.load(f))
            Jac.jac_import(
                target=mod,
                base_path=base,
                override_name="__main__" if main else None,
            )
    else:
        mach.close()
        raise ValueError("Not a valid file!\nOnly supports `.jac` and `.jir`")

    data = {}
    obj = Jac.get_object(id)
    if obj:
        data = obj.__jac__.__getstate__()
    else:
        print(f"Object with id {id} not found.", file=sys.stderr)
    mach.close()
    return data


@cmd_registry.register
def build(filename: str) -> None:
    """Build the specified .jac file.

    Compiles a Jac source file into a Jac Intermediate Representation (.jir) file,
    which can be executed more efficiently. Optionally performs type checking.

    Args:
        filename: Path to the .jac file to build
        typecheck: Perform type checking during build (default: True)

    Examples:
        jac build myprogram.jac
        jac build myprogram.jac --no-typecheck
    """
    if filename.endswith(".jac"):
        (out := JacProgram()).compile(file_path=filename)
        errs = len(out.errors_had)
        warnings = len(out.warnings_had)
        print(f"Errors: {errs}, Warnings: {warnings}")
        with open(filename[:-4] + ".jir", "wb") as f:
            pickle.dump(out, f)
    else:
        print("Not a .jac file.", file=sys.stderr)


@cmd_registry.register
def check(filename: str, print_errs: bool = True) -> None:
    """Run type checker for a specified .jac file.

    Performs static type analysis on a Jac program to identify potential type errors
    without executing the code. Useful for catching errors early in development.

    Args:
        filename: Path to the .jac file to check
        print_errs: Print detailed error messages (default: True)

    Examples:
        jac check myprogram.jac
        jac check myprogram.jac --no-print_errs
    """
    if filename.endswith(".jac"):
        (prog := JacProgram()).compile(file_path=filename)

        errs = len(prog.errors_had)
        warnings = len(prog.warnings_had)
        if print_errs:
            for e in prog.errors_had:
                print("Error:", e, file=sys.stderr)
        print(f"Errors: {errs}, Warnings: {warnings}")
    else:
        print("Not a .jac file.", file=sys.stderr)


@cmd_registry.register
def lsp() -> None:
    """Run Jac Language Server Protocol.

    Starts the Jac Language Server that provides IDE features like code completion,
    error checking, and navigation for Jac files. Used by editor extensions.

    Args:
        This command takes no parameters.

    Examples:
        jac lsp
    """
    from jaclang.langserve.server import run_lang_server

    run_lang_server()


@cmd_registry.register
def enter(
    filename: str,
    entrypoint: str,
    args: list,
    session: str = "",
    main: bool = True,
    root: str = "",
    node: str = "",
) -> None:
    """Run the specified entrypoint function in the given .jac file.

    Executes a specific function within a Jac program, allowing you to target
    particular functionality without running the entire program. Useful for
    testing specific components or running specific tasks.

    Args:
        filename: Path to the .jac or .jir file
        entrypoint: Name of the function to execute
        args: Arguments to pass to the entrypoint function
        session: Optional session identifier for persistent state
        main: Treat the module as __main__ (default: True)
        root: Root executor identifier
        node: Starting node identifier

    Examples:
        jac enter myprogram.jac main_function arg1 arg2
        jac enter myprogram.jac process_data --node data_node data.json
    """
    base, mod, mach = proc_file_sess(filename, session, root)

    if filename.endswith(".jac"):
        ret_module = Jac.jac_import(
            target=mod,
            base_path=base,
            override_name="__main__" if main else None,
        )
    elif filename.endswith(".jir"):
        with open(filename, "rb") as f:
            Jac.attach_program(pickle.load(f))
            ret_module = Jac.jac_import(
                target=mod,
                base_path=base,
                override_name="__main__" if main else None,
            )
    else:
        mach.close()
        raise ValueError("Not a valid file!\nOnly supports `.jac` and `.jir`")

    if ret_module:
        (loaded_mod,) = ret_module
        if not loaded_mod:
            print("Errors occurred while importing the module.", file=sys.stderr)
        else:
            archetype = getattr(loaded_mod, entrypoint)(*args)

            mach.set_entry_node(node)
            if isinstance(archetype, WalkerArchetype) and Jac.check_read_access(
                mach.entry_node
            ):
                Jac.spawn(mach.entry_node.archetype, archetype)
    mach.close()


@cmd_registry.register
def test(
    filepath: str,
    test_name: str = "",
    filter: str = "",
    xit: bool = False,
    maxfail: int = None,  # type:ignore
    directory: str = "",
    verbose: bool = False,
) -> None:
    """Run the test suite in the specified .jac file or directory.

    Executes test functions in Jac files to verify code correctness. Tests are
    identified by functions with names starting with 'test_'. Provides various
    options to control test execution and reporting.

    Args:
        filepath: Path to the .jac file or directory containing tests
        test_name: Run a specific test (without the 'test_' prefix)
        filter: Filter test files using Unix shell style patterns
        xit: Stop running tests as soon as an error is found
        maxfail: Stop running tests after specified number of failures
        directory: Run tests from the specified directory
        verbose: Show detailed test information and results

    Examples:
        jac test                     # Run all tests in current directory
        jac test mytest.jac          # Run all tests in mytest.jac
        jac test --test_name my_test # Run only test_my_test
        jac test --directory tests/  # Run all tests in tests/ directory
        jac test --filter "*_unit_*" # Run tests matching the pattern
        jac test --xit               # Stop on first failure
        jac test --verbose           # Show detailed output
    """
    failcount = Jac.run_test(
        filepath=filepath,
        func_name=("test_" + test_name) if test_name else None,
        filter=filter,
        xit=xit,
        maxfail=maxfail,
        directory=directory,
        verbose=verbose,
    )

    if failcount:
        raise SystemExit(f"Tests failed: {failcount}")


@cmd_registry.register
def tool(tool: str, args: Optional[list] = None) -> None:
    """Run the specified AST tool with optional arguments.

    Executes specialized tools for working with Jac's Abstract Syntax Tree (AST).
    These tools help with code analysis, transformation, and debugging.

    Args:
        tool: Name of the AST tool to run
        args: Optional arguments to pass to the tool

    Available Tools:
        list_tools: List all available AST tools

    Examples:
        jac tool list_tools
        jac tool <tool_name> [args...]
    """
    if hasattr(AstTool, tool):
        try:
            if args and len(args):
                print(getattr(AstTool(), tool)(args))
            else:
                print(getattr(AstTool(), tool)())
        except Exception as e:
            print(
                f"Error while running ast tool {tool}, check args: {e}", file=sys.stderr
            )
            raise e
    else:
        print(f"Ast tool {tool} not found.", file=sys.stderr)


@cmd_registry.register
def debug(filename: str, main: bool = True, cache: bool = False) -> None:
    """Debug the specified .jac file using the Python debugger.

    Runs a Jac program in debug mode, allowing you to set breakpoints, step through
    code execution, inspect variables, and troubleshoot issues interactively.

    Args:
        filename: Path to the .jac file to debug
        main: Treat the module as __main__ (default: True)
        cache: Use cached compilation if available (default: False)

    Examples:
        jac debug myprogram.jac

    Note:
        Add breakpoints in your code using the 'breakpoint()' function.
    """
    base, mod = os.path.split(filename)
    base = base if base else "./"
    mod = mod[:-4]
    if filename.endswith(".jac"):
        bytecode = JacProgram().compile(filename).gen.py_bytecode
        if bytecode:
            code = marshal.loads(bytecode)
            if db.has_breakpoint(bytecode):
                run(filename, main, cache)
            else:
                func = types.FunctionType(code, globals())

                print("Debugging with Jac debugger.\n")
                db.runcall(func)
                print("Done debugging.")
        else:
            print(f"Error while generating bytecode in {filename}.", file=sys.stderr)
    else:
        print("Not a .jac file.", file=sys.stderr)


@cmd_registry.register
def dot(
    filename: str,
    session: str = "",
    initial: str = "",
    depth: int = -1,
    traverse: bool = False,
    connection: list[str] = [],  # noqa: B006
    bfs: bool = False,
    edge_limit: int = 512,
    node_limit: int = 512,
    saveto: str = "",
) -> None:
    """Generate a DOT graph visualization from a Jac program.

    Creates a visual representation of the node graph in a Jac program using the
    DOT graph description language. The generated file can be visualized with
    tools like Graphviz.

    Args:
        filename: Path to the .jac file to visualize
        session: Optional session identifier for persistent state
        initial: Starting node for graph traversal (default: root node)
        depth: Maximum traversal depth (-1 for unlimited)
        traverse: Whether to traverse the graph structure (default: False)
        connection: List of edge types to include in the visualization
        bfs: Use breadth-first search for traversal (default: False)
        edge_limit: Maximum number of edges to include (default: 512)
        node_limit: Maximum number of nodes to include (default: 512)
        saveto: Output file path for the DOT file (default: <module_name>.dot)

    Examples:
        jac dot myprogram.jac
        jac dot myprogram.jac --initial root_node --depth 3
        jac dot myprogram.jac --traverse --connection edge_type1 edge_type2
        jac dot myprogram.jac --saveto graph.dot
    """
    base, mod, jac_machine = proc_file_sess(filename, session)

    if filename.endswith(".jac"):
        Jac.jac_import(target=mod, base_path=base, override_name="__main__")
        module = Jac.loaded_modules.get("__main__")
        globals().update(vars(module))
        try:
            node = globals().get(initial, eval(initial)) if initial else None
            graph = printgraph(
                node=node,
                depth=depth,
                traverse=traverse,
                edge_type=connection,
                bfs=bfs,
                edge_limit=edge_limit,
                node_limit=node_limit,
            )
        except Exception as e:
            print(f"Error while generating graph: {e}")
            import traceback

            traceback.print_exc()
            jac_machine.close()
            return
        file_name = saveto if saveto else f"{mod}.dot"
        with open(file_name, "w") as file:
            file.write(graph)
        print(f">>> Graph content saved to {os.path.join(os.getcwd(), file_name)}")
        jac_machine.close()
    else:
        print("Not a .jac file.", file=sys.stderr)


@cmd_registry.register
def py2jac(filename: str) -> None:
    """Convert a Python file to Jac code.

    Translates Python source code to equivalent Jac code, helping with migration
    from Python to Jac. The conversion handles basic syntax and structures but
    may require manual adjustments for complex code.

    Args:
        filename: Path to the .py file to convert

    Examples:
        jac py2jac myscript.py > converted.jac
    """
    if filename.endswith(".py"):
        with open(filename, "r") as f:
            file_source = f.read()
            code = PyastBuildPass(
                ir_in=uni.PythonModuleAst(
                    ast3.parse(file_source),
                    orig_src=uni.Source(file_source, filename),
                ),
                prog=JacProgram(),
            ).ir_out.unparse()
        print(code)
    else:
        print("Not a .py file.")


@cmd_registry.register
def jac2py(filename: str) -> None:
    """Convert a Jac file to Python code.

    Translates Jac source code to equivalent Python code. This is useful for
    understanding how Jac code is executed or for integrating Jac components
    with Python projects.

    Args:
        filename: Path to the .jac file to convert

    Examples:
        jac jac2py myprogram.jac > converted.py
    """
    if filename.endswith(".jac"):
        with open(filename, "r"):
            code = JacProgram().compile(file_path=filename).gen.py
        print(code)
    else:
        print("Not a .jac file.", file=sys.stderr)


def start_cli() -> None:
    """
    Start the command line interface.

    Returns:
    - None
    """
    parser = cmd_registry.parser
    args = parser.parse_args()
    cmd_registry.args = args

    if args.version:
        version = importlib.metadata.version("jaclang")
        print(f"Jac version {version}")
        print("Jac path:", __file__)
        return

    command = cmd_registry.get(args.command)
    if command:
        args_dict = vars(args)
        args_dict.pop("command")
        args_dict.pop("version", None)
        ret = command.call(**args_dict)
        if ret:
            print(ret)
    else:
        CommandShell(cmd_registry).cmdloop()


if __name__ == "__main__":
    start_cli()



"""Jac specific builtins."""

from __future__ import annotations

import json
from abc import abstractmethod
from enum import Enum
from typing import ClassVar, Optional, override

from jaclang.runtimelib.constructs import Archetype, NodeArchetype, Root
from jaclang.runtimelib.machine import JacMachineInterface as Jac


class AccessLevelEnum(Enum):
    """Access level constants for JAC objects."""

    NO_ACCESS = "NO_ACCESS"
    READ = "READ"
    CONNECT = "CONNECT"
    WRITE = "WRITE"


# Create module level constants for easier access
NoPerm = AccessLevelEnum.NO_ACCESS
ReadPerm = AccessLevelEnum.READ
ConnectPerm = AccessLevelEnum.CONNECT
WritePerm = AccessLevelEnum.WRITE


def printgraph(
    node: Optional[NodeArchetype] = None,
    depth: int = -1,
    traverse: bool = False,
    edge_type: Optional[list[str]] = None,
    bfs: bool = True,
    edge_limit: int = 512,
    node_limit: int = 512,
    file: Optional[str] = None,
    format: str = "dot",
) -> str:
    """Print the graph in different formats."""
    from jaclang.runtimelib.machine import JacMachineInterface as Jac

    fmt = format.lower()
    if fmt == "json":
        return _jac_graph_json(file)

    return Jac.printgraph(
        edge_type=edge_type,
        node=node or Jac.root(),
        depth=depth,
        traverse=traverse,
        bfs=bfs,
        edge_limit=edge_limit,
        node_limit=node_limit,
        file=file,
        format=fmt,
    )


def jid(obj: Archetype) -> str:
    """Get the id of the object."""
    return Jac.object_ref(obj)


def jobj(id: str) -> Archetype | None:
    """Get the object from the id."""
    return Jac.get_object(id)


def grant(obj: Archetype, level: AccessLevelEnum) -> None:
    """Grant permission for the object."""
    assert isinstance(level, AccessLevelEnum), f'Use {ConnectPerm} instead of "CONNECT"'
    Jac.perm_grant(obj, level=level.value)


def revoke(obj: Archetype) -> None:
    """Revoke permission for the object."""
    Jac.perm_revoke(obj)


def allroots() -> list[Jac.Root]:
    """Get all the roots."""
    return Jac.get_all_root()


def _jac_graph_json(file: Optional[str] = None) -> str:
    """Get the graph in json string."""
    processed: list[Root | NodeArchetype] = []
    nodes: list[dict] = []
    edges: list[dict] = []
    working_set: list[tuple] = []

    root = Jac.root()
    nodes.append({"id": id(root), "label": "root"})

    processed.append(root)
    working_set = [(root, ref) for ref in Jac.refs(root)]

    while working_set:
        start, end = working_set.pop(0)
        edges.append({"from": id(start), "to": id(end)})
        nodes.append({"id": id(end), "label": repr(end)})
        processed.append(end)
        for ref in Jac.refs(end):
            if ref not in processed:
                working_set.append((end, ref))
    output = json.dumps(
        {
            "version": "1.0",
            "nodes": nodes,
            "edges": edges,
        }
    )
    if file:
        with open(file, "w") as f:
            f.write(output)
    return output


__all__ = [
    "abstractmethod",
    "ClassVar",
    "override",
    "printgraph",
    "jid",
    "jobj",
    "grant",
    "revoke",
    "allroots",
    "NoPerm",
    "ReadPerm",
    "ConnectPerm",
    "WritePerm",
]



"""Semantic Definition Matching Pass for the Jac compiler.

This pass connects semantic string definitions (SemDefs) in the AST to their corresponding
declarations. It:

1. Establishes links between semantic string definitions in the main module and their declarations
2. Handles dotted name resolution for nested symbols
3. Issues warnings for unmatched semantic definitions

This pass is essential for Jac's semantic annotation system, allowing developers to define
semantics in one file and link them to their corresponding declarations in another.
"""

import jaclang.compiler.unitree as uni
from jaclang.compiler.passes.transform import Transform
from jaclang.compiler.unitree import Symbol, UniScopeNode


class SemDefMatchPass(Transform[uni.Module, uni.Module]):
    """Jac Semantic definition match pass."""

    def transform(self, ir_in: uni.Module) -> uni.Module:
        """Connect Decls and Defs."""
        self.cur_node = ir_in

        self.connect_sems(ir_in.sym_tab, ir_in.sym_tab)
        for impl_module in ir_in.impl_mod:
            self.connect_sems(impl_module.sym_tab, ir_in.sym_tab)

        return ir_in

    def find_symbol_from_dotted_name(
        self, dotted_name: str, target_sym_tab: UniScopeNode
    ) -> Symbol | None:
        """Find a symbol in the target symbol table by its dotted name.

        Example:
            "Foo.bar.baz" will do the following lookups:
            target_sym_tab.lookup("Foo").lookup("bar").lookup("baz")
        """
        parts = dotted_name.lstrip("sem.").split(".")
        current_sym_tab = target_sym_tab
        sym: uni.Symbol | None = None
        for part in parts:
            if current_sym_tab is None:
                return None
            sym = current_sym_tab.lookup(part, deep=False)
            if sym is None:
                return None
            current_sym_tab = sym.fetch_sym_tab
        return sym

    def connect_sems(
        self, source_sym_tab: UniScopeNode, target_sym_tab: UniScopeNode
    ) -> None:
        """Connect sem strings from source symbol table to declarations in target symbol table.

        When source_sym_tab and target_sym_tab are the same, this connects within the same module.
        When different, it connects implementations from impl_mod to declarations in the main module.
        """
        for sym in source_sym_tab.names_in_scope.values():
            if not isinstance(sym.decl.name_of, uni.SemDef):
                continue
            semdef = sym.decl.name_of
            target_sym = self.find_symbol_from_dotted_name(sym.sym_name, target_sym_tab)
            if target_sym is not None:
                target_sym.decl.name_of.semstr = semdef.value.lit_value
                target_sym.semstr = semdef.value.lit_value



from _typeshed import Incomplete

CODE_VERIFIER_PATTERN: Incomplete
CODE_CHALLENGE_PATTERN: Incomplete

def create_s256_code_challenge(code_verifier): ...
def compare_plain_code_challenge(code_verifier, code_challenge): ...
def compare_s256_code_challenge(code_verifier, code_challenge): ...

class CodeChallenge:
    DEFAULT_CODE_CHALLENGE_METHOD: str
    SUPPORTED_CODE_CHALLENGE_METHOD: Incomplete
    CODE_CHALLENGE_METHODS: Incomplete
    required: Incomplete
    def __init__(self, required: bool = True) -> None: ...
    def __call__(self, grant) -> None: ...
    def validate_code_challenge(self, grant) -> None: ...
    def validate_code_verifier(self, grant) -> None: ...
    def get_authorization_code_challenge(self, authorization_code): ...
    def get_authorization_code_challenge_method(self, authorization_code): ...



"""Test ast build pass module."""

import ast as ast3
import os
from difflib import unified_diff

import jaclang.compiler.unitree as uni
from jaclang.compiler.passes.tool import JacFormatPass
from jaclang.compiler.program import JacProgram
from jaclang.utils.helpers import add_line_numbers
from jaclang.utils.test import AstSyncTestMixin, TestCaseMicroSuite


class JacFormatPassTests(TestCaseMicroSuite):
    """Test pass module."""

    def compare_files(self, original_file: str, formatted_file: str = None) -> None:
        """Compare the original file with a provided formatted file or a new formatted version."""
        try:
            original_path = self.fixture_abs_path(original_file)
            with open(original_path, "r") as file:
                original_file_content = file.read()
            if formatted_file is None:
                formatted_content = JacProgram.jac_file_formatter(original_path)
            else:
                with open(self.fixture_abs_path(formatted_file), "r") as file:
                    formatted_content = file.read()
            diff = "\n".join(
                unified_diff(
                    original_file_content.splitlines(),
                    formatted_content.splitlines(),
                    fromfile="original",
                    tofile="formatted" if formatted_file is None else formatted_file,
                )
            )

            if diff:
                print(f"Differences found in comparison:\n{diff}")
                raise AssertionError("Files differ after formattinclearg.")

        except FileNotFoundError:
            print(f"File not found: {original_file} or {formatted_file}")
            raise
        except Exception as e:
            print(f"Error comparing files: {e}")
            raise

    def test_simple_walk_fmt(self) -> None:
        """Tests if the file matches a particular format."""
        self.compare_files(
            os.path.join(self.fixture_abs_path(""), "simple_walk_fmt.jac"),
        )

    def test_tagbreak(self) -> None:
        """Tests if the file matches a particular format."""
        self.compare_files(
            os.path.join(self.fixture_abs_path(""), "tagbreak.jac"),
        )

    def test_has_fmt(self) -> None:
        """Tests if the file matches a particular format."""
        self.compare_files(
            os.path.join(self.fixture_abs_path(""), "has_frmt.jac"),
        )

    def test_import_fmt(self) -> None:
        """Tests if the file matches a particular format."""
        self.compare_files(
            os.path.join(self.fixture_abs_path(""), "import_fmt.jac"),
        )

    def test_archetype(self) -> None:
        """Tests if the file matches a particular format."""
        self.compare_files(
            os.path.join(self.fixture_abs_path(""), "archetype_frmt.jac"),
        )

    # def test_corelib_fmt(self) -> None:
    #     """Tests if the file matches a particular format."""
    #     self.compare_files(
    #         os.path.join(self.fixture_abs_path(""), "corelib_fmt.jac"),
    #     )

    # def test_doc_string_fmt(self) -> None:
    #     """Tests if the file matches a particular format."""
    #     self.compare_files(
    #         os.path.join(self.fixture_abs_path(""), "doc_string.jac"),
    #     )

    # def test_compare_myca_fixtures(self) -> None:
    #     """Tests if files in the myca fixtures directory do not change after being formatted."""
    #     fixtures_dir = os.path.join(self.fixture_abs_path(""), "myca_formatted_code")
    #     fixture_files = os.listdir(fixtures_dir)
    #     for file_name in fixture_files:
    #         if file_name == "__jac_gen__":
    #             continue
    #         with self.subTest(file=file_name):
    #             file_path = os.path.join(fixtures_dir, file_name)
    #             self.compare_files(file_path)

    # def test_general_format_fixtures(self) -> None:
    #     """Tests if files in the general fixtures directory do not change after being formatted."""
    #     fixtures_dir = os.path.join(self.fixture_abs_path(""), "general_format_checks")
    #     fixture_files = os.listdir(fixtures_dir)
    #     for file_name in fixture_files:
    #         if file_name == "__jac_gen__":
    #             continue
    #         with self.subTest(file=file_name):
    #             file_path = os.path.join(fixtures_dir, file_name)
    #             self.compare_files(file_path)

    def micro_suite_test(self, filename: str) -> None:
        """
        Tests the Jac formatter by:
        1. Compiling a given Jac file.
        2. Formatting the Jac file content.
        3. Compiling the formatted content.
        4. Asserting that the AST of the original compilation and the
           AST of the formatted compilation are identical.
        This ensures that the formatting process does not alter the
        syntactic structure of the code.
        Includes a specific token check for 'circle_clean_tests.jac'.
        """
        code_gen_pure = JacProgram().compile(self.fixture_abs_path(filename))
        code_gen_format = JacProgram.jac_file_formatter(self.fixture_abs_path(filename))
        code_gen_jac = JacProgram().compile(use_str=code_gen_format, file_path=filename)
        if "circle_clean_tests.jac" in filename:
            tokens = code_gen_format.split()
            num_test = 0
            for i in range(len(tokens)):
                if tokens[i] == "test":
                    num_test += 1
                    self.assertEqual(tokens[i + 1], "{")
            self.assertEqual(num_test, 3)
            return
        try:
            before = ast3.dump(code_gen_pure.gen.py_ast[0], indent=2)
            after = ast3.dump(code_gen_jac.gen.py_ast[0], indent=2)
            self.assertTrue(
                isinstance(code_gen_pure, uni.Module)
                and isinstance(code_gen_jac, uni.Module),
                "Parsed objects are not modules.",
            )

            diff = "\n".join(unified_diff(before.splitlines(), after.splitlines()))
            self.assertFalse(diff, "AST structures differ after formatting.")

        except Exception as e:
            print(f"Error in {filename}: {e}")
            print(add_line_numbers(code_gen_pure.source.code))
            print("\n+++++++++++++++++++++++++++++++++++++++\n")
            print(add_line_numbers(code_gen_format))
            print("\n+++++++++++++++++++++++++++++++++++++++\n")
            print("\n".join(unified_diff(before.splitlines(), after.splitlines())))
            raise e


JacFormatPassTests.self_attach_micro_tests()



from networkx.utils.backends import _dispatchable

def write_p2g(G, path, encoding: str = "utf-8") -> None: ...
@_dispatchable
def read_p2g(path, encoding: str = "utf-8"): ...
@_dispatchable
def parse_p2g(lines): ...
